from time import perf_counter
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram
from pymemcache.client.base import PooledClient

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.lib import config
from baseplate.lib import metrics
from baseplate.lib.prometheus_metrics import default_latency_buckets


Serializer = Callable[[str, Any], Tuple[bytes, int]]
Deserializer = Callable[[str, bytes, int], Any]


def pool_from_config(
    app_config: config.RawConfig,
    prefix: str = "memcache.",
    serializer: Optional[Serializer] = None,
    deserializer: Optional[Deserializer] = None,
) -> PooledClient:
    """Make a PooledClient from a configuration dictionary.

    The keys useful to :py:func:`pool_from_config` should be prefixed, e.g.
    ``memcache.endpoint``, ``memcache.max_pool_size``, etc. The ``prefix``
    argument specifies the prefix used to filter keys. Each key is mapped to a
    corresponding keyword argument on the
    :py:class:`~pymemcache.client.base.PooledClient` constructor.

    Supported keys:

    * ``endpoint`` (required): a string representing a host and port to connect
        to memcached service, e.g. ``localhost:11211`` or ``127.0.0.1:11211``.
    * ``max_pool_size``: an integer for the maximum pool size to use, by default
        this is ``2147483648``.
    * ``connect_timeout``: how long (as
        :py:func:`~baseplate.lib.config.Timespan`) to wait for a connection to
        memcached server. Defaults to the underlying socket default timeout.
    * ``timeout``: how long (as :py:func:`~baseplate.lib.config.Timespan`) to
        wait for calls on the socket connected to memcache. Defaults to the
        underlying socket default timeout.

    :param app_config: the raw application configuration
    :param prefix: prefix for configuration keys
    :param serializer: function to serialize values to strings suitable
        for being stored in memcached. An example is
        :py:func:`~baseplate.clients.memcache.lib.make_dump_and_compress_fn`.
    :param deserializer: function to convert strings returned from
        memcached to arbitrary objects, must be compatible with ``serializer``.
        An example is :py:func:`~baseplate.clients.memcache.lib.decompress_and_load`.

    :returns: :py:class:`pymemcache.client.base.PooledClient`

    """
    assert prefix.endswith(".")
    parser = config.SpecParser(
        {
            "endpoint": config.Endpoint,
            "max_pool_size": config.Optional(config.Integer, default=None),
            "connect_timeout": config.Optional(config.TimespanWithLegacyFallback, default=None),
            "timeout": config.Optional(config.TimespanWithLegacyFallback, default=None),
            "no_delay": config.Optional(config.Boolean, default=True),
        }
    )
    options = parser.parse(prefix[:-1], app_config)

    return PooledClient(
        server=options.endpoint.address,
        connect_timeout=options.connect_timeout and options.connect_timeout.total_seconds(),
        timeout=options.timeout and options.timeout.total_seconds(),
        serializer=serializer,
        deserializer=deserializer,
        no_delay=options.no_delay,
        max_pool_size=options.max_pool_size,
    )


class MemcacheClient(config.Parser):
    """Configure a Memcached client.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    See :py:func:`pool_from_config` for available configuration settings.

    :param serializer: function to serialize values to strings suitable
        for being stored in memcached. An example is
        :py:func:`~baseplate.clients.memcache.lib.make_dump_and_compress_fn`.
    :param deserializer: function to convert strings returned from
        memcached to arbitrary objects, must be compatible with ``serializer``.
        An example is :py:func:`~baseplate.clients.memcache.lib.decompress_and_load`.

    """

    def __init__(
        self, serializer: Optional[Serializer] = None, deserializer: Optional[Deserializer] = None
    ):
        self.serializer = serializer
        self.deserializer = deserializer

    def parse(self, key_path: str, raw_config: config.RawConfig) -> "MemcacheContextFactory":
        pool = pool_from_config(
            raw_config,
            prefix=f"{key_path}.",
            serializer=self.serializer,
            deserializer=self.deserializer,
        )
        return MemcacheContextFactory(pool, key_path)


class MemcacheContextFactory(ContextFactory):
    """Memcache client context factory.

    This factory will attach a
    :py:class:`~baseplate.clients.memcache.MonitoredMemcacheConnection` to an
    attribute on the :py:class:`~baseplate.RequestContext`. When memcache
    commands are executed via this connection object, they will use connections
    from the provided :py:class:`~pymemcache.client.base.PooledClient` and
    automatically record diagnostic information.

    :param pooled_client: A pooled client.

    """

    PROM_PREFIX = "memcached_client_pool"
    PROM_LABELS = ["memcached_pool"]

    pool_size_gauge = Gauge(
        f"{PROM_PREFIX}_max_size",
        "Maximum number of connections allowed in this pool",
        PROM_LABELS,
    )

    used_connections_gauge = Gauge(
        f"{PROM_PREFIX}_active_connections",
        "Number of connections in this pool currently in use",
        PROM_LABELS,
    )

    free_connections_gauge = Gauge(
        f"{PROM_PREFIX}_idle_connections",
        "Number of free connections in this pool",
        PROM_LABELS,
    )

    def __init__(self, pooled_client: PooledClient, name: str = "default"):
        self.pooled_client = pooled_client
        self.name = name

    def report_memcache_runtime_metrics(self, batch: metrics.Client) -> None:
        pool = self.pooled_client.client_pool
        self.pool_size_gauge.labels(self.name).set(pool.max_size)
        self.free_connections_gauge.labels(self.name).set(len(pool.free))
        self.used_connections_gauge.labels(self.name).set(len(pool.used))

        batch.gauge("pool.in_use").replace(len(pool.used))
        batch.gauge("pool.open_and_available").replace(len(pool.free))
        batch.gauge("pool.size").replace(pool.max_size)

    def make_object_for_context(self, name: str, span: Span) -> "MonitoredMemcacheConnection":
        return MonitoredMemcacheConnection(name, span, self.pooled_client)


Key = Union[str, bytes]

PROM_NAMESPACE = "memcached"
LABELS_COMMON = [
    f"{PROM_NAMESPACE}_address",
    f"{PROM_NAMESPACE}_command",
]
LATENCY_SECONDS = Histogram(
    f"{PROM_NAMESPACE}_client_latency_seconds",
    "Latency histogram of outoing memcached requests",
    [*LABELS_COMMON, f"{PROM_NAMESPACE}_success"],
    buckets=default_latency_buckets,
)
REQUESTS_TOTAL = Counter(
    f"{PROM_NAMESPACE}_client_requests_total",
    "Total number of memcached requests",
    [*LABELS_COMMON, f"{PROM_NAMESPACE}_success"],
)
ACTIVE_REQUESTS = Gauge(
    f"{PROM_NAMESPACE}_client_active_requests",
    "Number of active requests",
    LABELS_COMMON,
)


def _prom_instrument(func: Any) -> Any:
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        labels_common = {
            f"{PROM_NAMESPACE}_address": str(self.pooled_client.server),
            f"{PROM_NAMESPACE}_command": func.__name__,
        }
        success = "true"
        start_time = perf_counter()

        try:
            with ACTIVE_REQUESTS.labels(**labels_common).track_inprogress():
                return func(self, *args, **kwargs)
        except:  # noqa
            success = "false"
            raise
        finally:
            REQUESTS_TOTAL.labels(**{**labels_common, f"{PROM_NAMESPACE}_success": success}).inc()
            LATENCY_SECONDS.labels(
                **{**labels_common, f"{PROM_NAMESPACE}_success": success}
            ).observe(perf_counter() - start_time)

    return wrapper


class MonitoredMemcacheConnection:
    """Memcache connection that collects diagnostic information.

    This connection acts like a
    :py:class:`~pymemcache.client.base.PooledClient` except that operations are
    wrapped with diagnostic collection. Some methods may not yet be wrapped
    with monitoring. Please request assistance if any needed methods are not
    being monitored.

    """

    def __init__(self, context_name: str, server_span: Span, pooled_client: PooledClient):
        self.context_name = context_name
        self.server_span = server_span
        self.pooled_client = pooled_client

    @_prom_instrument
    def close(self) -> None:
        with self._make_span("close"):
            return self.pooled_client.close()

    @_prom_instrument
    def set(self, key: Key, value: Any, expire: int = 0, noreply: Optional[bool] = None) -> bool:
        with self._make_span("set") as span:
            span.set_tag("key", key)
            span.set_tag("expire", expire)
            span.set_tag("noreply", noreply)
            return self.pooled_client.set(key, value, expire=expire, noreply=noreply)

    @_prom_instrument
    def set_many(
        self, values: Dict[Key, Any], expire: int = 0, noreply: Optional[bool] = None
    ) -> List[str]:
        with self._make_span("set_many") as span:
            span.set_tag("key_count", len(values))
            span.set_tag("keys", make_keys_str(values.keys()))
            span.set_tag("expire", expire)
            span.set_tag("noreply", noreply)
            return self.pooled_client.set_many(values, expire=expire, noreply=noreply)

    @_prom_instrument
    def replace(
        self, key: Key, value: Any, expire: int = 0, noreply: Optional[bool] = None
    ) -> bool:
        with self._make_span("replace") as span:
            span.set_tag("key", key)
            span.set_tag("expire", expire)
            span.set_tag("noreply", noreply)
            return self.pooled_client.replace(key, value, expire=expire, noreply=noreply)

    @_prom_instrument
    def append(self, key: Key, value: Any, expire: int = 0, noreply: Optional[bool] = None) -> bool:
        with self._make_span("append") as span:
            span.set_tag("key", key)
            span.set_tag("expire", expire)
            span.set_tag("noreply", noreply)
            return self.pooled_client.append(key, value, expire=expire, noreply=noreply)

    @_prom_instrument
    def prepend(
        self, key: Key, value: Any, expire: int = 0, noreply: Optional[bool] = None
    ) -> bool:
        with self._make_span("prepend") as span:
            span.set_tag("key", key)
            span.set_tag("expire", expire)
            span.set_tag("noreply", noreply)
            return self.pooled_client.prepend(key, value, expire=expire, noreply=noreply)

    @_prom_instrument
    def cas(
        self, key: Key, value: Any, cas: int, expire: int = 0, noreply: Optional[bool] = None
    ) -> Optional[bool]:
        with self._make_span("cas") as span:
            span.set_tag("key", key)
            span.set_tag("cas", cas)
            span.set_tag("expire", expire)
            span.set_tag("noreply", noreply)
            return self.pooled_client.cas(key, value, cas, expire=expire, noreply=noreply)

    @_prom_instrument
    def get(self, key: Key, default: Any = None) -> Any:
        with self._make_span("get") as span:
            span.set_tag("key", key)
            kwargs = {}
            if default is not None:
                kwargs["default"] = default
            return self.pooled_client.get(key, **kwargs)

    @_prom_instrument
    def get_many(self, keys: Sequence[Key]) -> Dict[Key, Any]:
        with self._make_span("get_many") as span:
            span.set_tag("key_count", len(keys))
            span.set_tag("keys", make_keys_str(keys))
            return self.pooled_client.get_many(keys)

    @_prom_instrument
    def gets(
        self, key: Key, default: Optional[Any] = None, cas_default: Optional[Any] = None
    ) -> Tuple[Any, Any]:
        with self._make_span("gets") as span:
            span.set_tag("key", key)
            return self.pooled_client.gets(key, default=default, cas_default=cas_default)

    @_prom_instrument
    def gets_many(self, keys: Sequence[Key]) -> Dict[Key, Tuple[Any, Any]]:
        with self._make_span("gets_many") as span:
            span.set_tag("key_count", len(keys))
            span.set_tag("keys", make_keys_str(keys))
            return self.pooled_client.gets_many(keys)

    @_prom_instrument
    def delete(self, key: Key, noreply: Optional[bool] = None) -> bool:
        with self._make_span("delete") as span:
            span.set_tag("key", key)
            span.set_tag("noreply", noreply)
            return self.pooled_client.delete(key, noreply=noreply)

    @_prom_instrument
    def delete_many(self, keys: Sequence[Key], noreply: Optional[bool] = None) -> bool:
        with self._make_span("delete_many") as span:
            span.set_tag("key_count", len(keys))
            span.set_tag("noreply", noreply)
            span.set_tag("keys", make_keys_str(keys))
            return self.pooled_client.delete_many(keys, noreply=noreply)

    @_prom_instrument
    def add(self, key: Key, value: Any, expire: int = 0, noreply: Optional[bool] = None) -> bool:
        with self._make_span("add") as span:
            span.set_tag("key", key)
            span.set_tag("expire", expire)
            span.set_tag("noreply", noreply)
            return self.pooled_client.add(key, value, expire=expire, noreply=noreply)

    @_prom_instrument
    def incr(self, key: Key, value: int, noreply: Optional[bool] = False) -> Optional[int]:
        with self._make_span("incr") as span:
            span.set_tag("key", key)
            span.set_tag("noreply", noreply)
            return self.pooled_client.incr(key, value, noreply=noreply)

    @_prom_instrument
    def decr(self, key: Key, value: int, noreply: Optional[bool] = False) -> Optional[int]:
        with self._make_span("decr") as span:
            span.set_tag("key", key)
            span.set_tag("noreply", noreply)
            return self.pooled_client.decr(key, value, noreply=noreply)

    @_prom_instrument
    def touch(self, key: Key, expire: int = 0, noreply: Optional[bool] = None) -> bool:
        with self._make_span("touch") as span:
            span.set_tag("key", key)
            span.set_tag("expire", expire)
            span.set_tag("noreply", noreply)
            return self.pooled_client.touch(key, expire=expire, noreply=noreply)

    @_prom_instrument
    def stats(self, *args: str) -> Dict[str, Any]:
        with self._make_span("stats"):
            return self.pooled_client.stats(*args)

    @_prom_instrument
    def flush_all(self, delay: int = 0, noreply: Optional[bool] = None) -> bool:
        with self._make_span("flush_all") as span:
            span.set_tag("delay", delay)
            span.set_tag("noreply", noreply)
            return self.pooled_client.flush_all(delay=delay, noreply=noreply)

    @_prom_instrument
    def quit(self) -> None:
        with self._make_span("quit"):
            return self.pooled_client.quit()

    def _make_span(self, method_name: str) -> Span:
        """Get a child span of the current server span.

        The returned span is tagged with ``method_name`` and given a name
        that corresponds to the current context name and called method.

        """
        trace_name = f"{self.context_name}.{method_name}"
        span = self.server_span.make_child(trace_name)
        span.set_tag("method", method_name)
        return span


def make_keys_str(keys: Iterable[Key]) -> str:
    """Make a string representation of an iterable of keys."""
    keys_str = ",".join(x.decode("utf-8") if isinstance(x, bytes) else x for x in keys)
    if len(keys_str) > 100:
        return keys_str[:100] + "..."
    return keys_str



import unittest

from unittest import mock

import requests

from baseplate.lib import metrics
from baseplate.sidecars import SerializedBatch
from baseplate.sidecars import trace_publisher


class ZipkinPublisherTest(unittest.TestCase):
    @mock.patch("requests.Session", autospec=True)
    def setUp(self, mock_Session):
        self.session = mock_Session.return_value
        self.session.headers = {}
        self.metrics_client = mock.MagicMock(autospec=metrics.Client)
        self.zipkin_api_url = "http://test.local/api/v2"
        self.publisher = trace_publisher.ZipkinPublisher(self.zipkin_api_url, self.metrics_client)

    def test_initialization(self):
        self.assertEqual(self.publisher.endpoint, f"{self.zipkin_api_url}/spans")
        self.publisher.session.mount.assert_called_with("http://", mock.ANY)

    def test_empty_batch(self):
        self.publisher.publish(SerializedBatch(item_count=0, serialized=b""))
        self.assertEqual(self.session.post.call_count, 0)

    def test_publish_retry(self):
        # raise two errors and then return a mock response
        self.session.post.side_effect = [requests.HTTPError(504), OSError, mock.Mock()]
        spans = b"[]"
        self.publisher.publish(SerializedBatch(item_count=1, serialized=spans))
        self.assertEqual(self.session.post.call_count, 3)



import contextlib
import time
import unittest

import posix_ipc

from baseplate.lib.message_queue import MessageQueue
from baseplate.lib.message_queue import TimedOutError


class TestMessageQueueCreation(unittest.TestCase):
    qname = "/baseplate-test-queue"

    def setUp(self):
        try:
            queue = posix_ipc.MessageQueue(self.qname)
        except posix_ipc.ExistentialError:
            pass
        else:
            queue.unlink()
            queue.close()

    def test_create_queue(self):
        message_queue = MessageQueue(self.qname, max_messages=1, max_message_size=1)

        with contextlib.closing(message_queue) as mq:
            self.assertEqual(mq.queue.max_messages, 1)
            self.assertEqual(mq.queue.max_message_size, 1)

    def test_put_get(self):
        message_queue = MessageQueue(self.qname, max_messages=1, max_message_size=1)

        with contextlib.closing(message_queue) as mq:
            mq.put(b"x")
            message = mq.get()
            self.assertEqual(message, b"x")

    def test_get_timeout(self):
        message_queue = MessageQueue(self.qname, max_messages=1, max_message_size=1)

        with contextlib.closing(message_queue) as mq:
            start = time.time()
            with self.assertRaises(TimedOutError):
                mq.get(timeout=0.1)
            elapsed = time.time() - start
            self.assertAlmostEqual(elapsed, 0.1, places=2)

    def test_put_timeout(self):
        message_queue = MessageQueue(self.qname, max_messages=1, max_message_size=1)

        with contextlib.closing(message_queue) as mq:
            mq.put(b"x")
            start = time.time()
            with self.assertRaises(TimedOutError):
                mq.put(b"x", timeout=0.1)
            elapsed = time.time() - start
            self.assertAlmostEqual(elapsed, 0.1, places=1)

    def test_put_zero_timeout(self):
        message_queue = MessageQueue(self.qname, max_messages=1, max_message_size=1)

        with contextlib.closing(message_queue) as mq:
            mq.put(b"x", timeout=0)
            message = mq.get()
            self.assertEqual(message, b"x")

    def test_put_full_zero_timeout(self):
        message_queue = MessageQueue(self.qname, max_messages=1, max_message_size=1)

        with contextlib.closing(message_queue) as mq:
            mq.put(b"1", timeout=0)

            with self.assertRaises(TimedOutError):
                mq.put(b"2", timeout=0)

    def tearDown(self):
        try:
            queue = posix_ipc.MessageQueue(self.qname)
        except posix_ipc.ExistentialError:
            pass
        else:
            queue.unlink()
            queue.close()



import time
import unittest

from unittest import mock

import gevent.socket

try:
    from kazoo.exceptions import NoNodeError
    from kazoo.handlers.gevent import SequentialGeventHandler
    from kazoo.handlers.threading import SequentialThreadingHandler
    from kazoo.retry import KazooRetry
except ImportError:
    raise unittest.SkipTest("kazoo is not installed")

from baseplate.lib.live_data.zookeeper import zookeeper_client_from_config
from baseplate.lib.secrets import SecretsStore

from .. import get_endpoint_or_skip_container

zookeeper_endpoint = get_endpoint_or_skip_container("zookeeper", 2181)


class ZooKeeperHandlerWithPatchingTests(unittest.TestCase):
    def test_create_client_no_secrets(self):
        secrets = mock.Mock(spec=SecretsStore)
        client = zookeeper_client_from_config(
            secrets, {"zookeeper.hosts": "%s:%d" % zookeeper_endpoint.address}
        )

        client.start()

        with self.assertRaises(NoNodeError):
            client.get("/does_not_exist")

        client.stop()

    def test_create_client_with_credentials(self):
        secrets = mock.Mock(spec=SecretsStore)
        secrets.get_simple.return_value = b"myzkuser:hunter2"

        client = zookeeper_client_from_config(
            secrets,
            {
                "zookeeper.hosts": "%s:%d" % zookeeper_endpoint.address,
                "zookeeper.credentials": "secret/zk-user",
            },
        )

        client.start()
        with self.assertRaises(NoNodeError):
            client.get("/does_not_exist")
        client.stop()

        secrets.get_simple.assert_called_with("secret/zk-user")
        self.assertEqual(list(client.auth_data), [("digest", "myzkuser:hunter2")])

    def test_create_client_uses_threading_handler_when_not_gevent_patched(self):
        secrets = mock.Mock(spec=SecretsStore)
        client = zookeeper_client_from_config(
            secrets, {"zookeeper.hosts": "%s:%d" % zookeeper_endpoint.address}
        )
        assert isinstance(client.handler, SequentialThreadingHandler)

    @mock.patch("baseplate.lib.live_data.zookeeper.gevent_is_patched", return_value=True)
    def test_create_client_uses_gevent_handler_when_gevent_patched(self, mock_gevent_is_patched):
        patched_default_values = tuple(
            gevent.sleep if value is time.sleep else value
            for value in KazooRetry.__init__.__defaults__
        )
        with mock.patch("kazoo.retry.KazooRetry.__init__.__defaults__", patched_default_values):
            secrets = mock.Mock(spec=SecretsStore)
            client = zookeeper_client_from_config(
                secrets, {"zookeeper.hosts": "%s:%d" % zookeeper_endpoint.address}
            )
            assert isinstance(client.handler, SequentialGeventHandler)



import unittest

from unittest import mock

import requests

from baseplate.lib import config
from baseplate.lib import metrics
from baseplate.sidecars import event_publisher
from baseplate.sidecars import SerializedBatch


class TimeLimitedBatchTests(unittest.TestCase):
    def setUp(self):
        self.inner = mock.Mock(autospec=event_publisher.Batch)
        self.batch = event_publisher.TimeLimitedBatch(self.inner, max_age=1)

    def test_serialize(self):
        result = self.batch.serialize()
        self.assertEqual(result, self.inner.serialize.return_value)

    @mock.patch("time.time")
    def test_add_after_time_up_and_reset(self, mock_time):
        mock_time.return_value = 1

        self.batch.add("a")
        self.assertEqual(self.inner.add.call_count, 1)

        with self.assertRaises(event_publisher.BatchFull):
            mock_time.return_value = 3
            self.batch.add("b")
        self.assertEqual(self.inner.add.call_count, 1)

        self.batch.reset()
        self.batch.add("b")
        self.assertEqual(self.inner.add.call_count, 2)
        self.assertEqual(self.inner.reset.call_count, 1)


class BatchTests(unittest.TestCase):
    def test_v2(self):
        batch = event_publisher.V2Batch(max_size=50)
        batch.add(None)
        batch.add(b"a")
        batch.add(b"b")

        result = batch.serialize()
        self.assertEqual(result.item_count, 2)
        self.assertEqual(result.serialized, b'{"1":{"lst":["rec",2,a,b]}}')

        with self.assertRaises(event_publisher.BatchFull):
            batch.add(b"x" * 100)

        batch.reset()
        result = batch.serialize()
        self.assertEqual(result.item_count, 0)


class V2JBatchTests(unittest.TestCase):
    def test_v2j(self):
        batch = event_publisher.V2JBatch(max_size=50)
        batch.add(None)
        batch.add(b'{"source":1}')
        batch.add(b'{"source":2}')

        result = batch.serialize()
        self.assertEqual(result.item_count, 2)
        self.assertEqual(result.serialized, b'[{"source":1},{"source":2}]')

        with self.assertRaises(event_publisher.BatchFull):
            batch.add(b"x" * 100)

        batch.reset()
        result = batch.serialize()
        self.assertEqual(result.item_count, 0)


class PublisherTests(unittest.TestCase):
    @mock.patch("requests.Session", autospec=True)
    def setUp(self, Session):
        self.config = config.ConfigNamespace()
        self.config.collector = config.ConfigNamespace()
        self.config.collector.hostname = "test.local"
        self.config.collector.version = 1
        self.config.collector.scheme = "https"
        self.config.key = config.ConfigNamespace()
        self.config.key.name = "TestKey"
        self.config.key.secret = b"hunter2"

        self.session = Session.return_value
        self.session.headers = {}

        self.metrics_client = mock.MagicMock(autospec=metrics.Client)

        self.publisher = event_publisher.BatchPublisher(self.metrics_client, self.config)

    def test_empty_batch(self):
        self.publisher.publish(SerializedBatch(item_count=0, serialized=b""))
        self.assertEqual(self.session.post.call_count, 0)

    def test_publish_success(self):
        events = b'[{"example": "value"}]'

        self.publisher.publish(SerializedBatch(item_count=1, serialized=events))

        self.assertEqual(self.session.post.call_count, 1)

        args, kwargs = self.session.post.call_args
        headers = kwargs.get("headers", {})
        self.assertEqual(args[0], "https://test.local/v1")
        self.assertEqual(headers["Content-Type"], "application/json")
        self.assertEqual(
            headers["X-Signature"],
            "key=TestKey, mac=7c46d56b99cd4cb05e08238c1d4c10a2f330795e9d7327f17cc66fd206bf1179",
        )

    @mock.patch("time.sleep")
    def test_fail_on_client_error(self, mock_sleep):
        self.session.post.side_effect = [
            requests.HTTPError(400, response=mock.Mock(status_code=400))
        ]
        events = b'[{"example": "value"}]'

        with self.assertRaises(requests.HTTPError):
            self.publisher.publish(SerializedBatch(item_count=1, serialized=events))

        self.assertEqual(mock_sleep.call_count, 0)
        self.assertEqual(self.session.post.call_count, 1)



import unittest

from datetime import datetime
from datetime import timezone

import pytz

from baseplate.lib.datetime import datetime_to_epoch_milliseconds
from baseplate.lib.datetime import datetime_to_epoch_seconds
from baseplate.lib.datetime import epoch_milliseconds_to_datetime
from baseplate.lib.datetime import epoch_seconds_to_datetime
from baseplate.lib.datetime import get_utc_now


EXAMPLE_DATETIME = datetime.utcnow().replace(tzinfo=timezone.utc, microsecond=0)


class DatetimeTests(unittest.TestCase):
    def test_datetime_conversions(self):
        epoch_sec = datetime_to_epoch_seconds(EXAMPLE_DATETIME)
        epoch_ms = datetime_to_epoch_milliseconds(EXAMPLE_DATETIME)
        self.assertEqual(EXAMPLE_DATETIME, epoch_seconds_to_datetime(epoch_sec))
        self.assertEqual(epoch_sec, int(epoch_ms / 1000))
        self.assertEqual(EXAMPLE_DATETIME, epoch_milliseconds_to_datetime(epoch_ms))

    def test_timezone_equivalence(self):
        pytz_datetime = EXAMPLE_DATETIME.replace(tzinfo=pytz.UTC)
        self.assertEqual(
            datetime_to_epoch_milliseconds(pytz_datetime),
            datetime_to_epoch_milliseconds(EXAMPLE_DATETIME),
        )
        self.assertEqual(
            datetime_to_epoch_seconds(pytz_datetime), datetime_to_epoch_seconds(EXAMPLE_DATETIME)
        )

    def test_get_utc_now(self):
        now = get_utc_now()
        self.assertIsNotNone(now.tzinfo)



#
# Autogenerated by Thrift Compiler (0.14.1)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:slots
#
import sys

from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.Thrift import TFrozenDict
from thrift.Thrift import TMessageType
from thrift.Thrift import TType
from thrift.TRecursive import fix_spec

from .ttypes import *



# pylint: disable=wrong-import-position,wrong-import-order
from gevent.monkey import patch_all

from baseplate.server.monkey import patch_stdlib_queues

# In order to allow Prometheus to scrape metrics, we need to concurrently
# handle requests to '/metrics' along with the sidecar's execution.
# Monkey patching is used to replace the stdlib sequential versions of functions
# with concurrent versions. It must happen as soon as possible, before the
# sequential versions are imported.
patch_all()
patch_stdlib_queues()

import argparse
import configparser
import email.utils
import gzip
import hashlib
import hmac
import logging

from typing import Any
from typing import List
from typing import Optional

import gevent
from prometheus_client import Counter, Histogram
import requests

from baseplate import __version__ as baseplate_version
from baseplate.lib import config
from baseplate.lib import metrics
from baseplate.lib.events import MAX_EVENT_SIZE
from baseplate.lib.events import MAX_QUEUE_SIZE
from baseplate.lib.message_queue import MessageQueue
from baseplate.lib.message_queue import TimedOutError
from baseplate.lib.metrics import metrics_client_from_config
from baseplate.lib.retry import RetryPolicy
from baseplate.server import EnvironmentInterpolation
from baseplate.server.prometheus import start_prometheus_exporter_for_sidecar
from baseplate.sidecars import Batch
from baseplate.sidecars import BatchFull
from baseplate.sidecars import SerializedBatch
from baseplate.sidecars import TimeLimitedBatch


logger = logging.getLogger(__name__)


# seconds to wait for the event collector
POST_TIMEOUT = 3
# Base rate for expontential retry delay
RETRY_BACKOFF = 2
# Maximum wait for a message to be sent
MAX_RETRY_TIME = 5 * 60
# maximum time (seconds) an event can sit around while we wait for more
# messages to batch
MAX_BATCH_AGE = 1
# maximum size (in bytes) of a batch of events
MAX_BATCH_SIZE = 500 * 1024


class MaxRetriesError(Exception):
    pass


class V2Batch(Batch):
    # V2 batches are a struct with a single field of list<Event> type. because
    # we don't have the individual event schemas here, but pre-serialized
    # individual events instead, we mimic TJSONProtocol's container format
    # manually here.
    #
    # the format string in this header is for the length field. it should be
    # kept a similar number of bytes to the output of the formatting so the
    # initial size estimate is OK. 4 digits should be plenty.
    _header = '{"1":{"lst":["rec",%01d,'
    _end = b"]}}"

    def __init__(self, max_size: int = MAX_BATCH_SIZE):
        self.max_size = max_size
        self.reset()

    def add(self, item: Optional[bytes]) -> None:
        if not item:
            return

        serialized_size = len(item) + 1  # the comma at the end

        if self._size + serialized_size > self.max_size:
            raise BatchFull

        self._items.append(item)
        self._size += serialized_size

    def serialize(self) -> SerializedBatch:
        header = (self._header % len(self._items)).encode()
        return SerializedBatch(
            item_count=len(self._items), serialized=header + b",".join(self._items) + self._end
        )

    def reset(self) -> None:
        self._items: List[bytes] = []
        self._size = len(self._header) + len(self._end)


class V2JBatch(V2Batch):
    # Send a batch as a plain JSON array.  Useful when your events are not
    # Thrift JSON
    _header = "["
    _end = b"]"

    def serialize(self) -> SerializedBatch:
        serialized = self._header.encode() + b",".join(self._items) + self._end
        return SerializedBatch(item_count=len(self._items), serialized=serialized)


publishCountTotal = Counter("event_publish_total", "total count of published events")
publishLatency = Histogram("event_publish_latency", "latency for publishing a batch of events")
publishErrors = Counter("event_publish_errors", "total count of published events", ["error_type"])


class BatchPublisher:
    def __init__(self, metrics_client: metrics.Client, cfg: Any):
        self.metrics = metrics_client
        self.url = f"{cfg.collector.scheme}://{cfg.collector.hostname}/v{cfg.collector.version}"
        self.key_name = cfg.key.name
        self.key_secret = cfg.key.secret
        self.session = requests.Session()
        self.session.headers[
            "User-Agent"
        ] = f"baseplate.py-{self.__class__.__name__}/{baseplate_version}"

    def _sign_payload(self, payload: bytes) -> str:
        digest = hmac.new(self.key_secret, payload, hashlib.sha256).hexdigest()
        return f"key={self.key_name}, mac={digest}"

    def publish(self, payload: SerializedBatch) -> None:
        if not payload.item_count:
            return

        logger.info("sending batch of %d events", payload.item_count)
        compressed_payload = gzip.compress(payload.serialized)
        headers = {
            "Date": email.utils.formatdate(usegmt=True),
            "User-Agent": "baseplate-event-publisher/1.0",
            "Content-Type": "application/json",
            "X-Signature": self._sign_payload(payload.serialized),
            "Content-Encoding": "gzip",
        }

        for _ in RetryPolicy.new(budget=MAX_RETRY_TIME, backoff=RETRY_BACKOFF):
            try:
                with self.metrics.timer("post"):
                    with publishLatency.time():
                        response = self.session.post(
                            self.url,
                            headers=headers,
                            data=compressed_payload,
                            timeout=POST_TIMEOUT,
                            # http://docs.python-requests.org/en/latest/user/advanced/#keep-alive
                            stream=False,
                        )
                response.raise_for_status()
            except requests.HTTPError as exc:
                publishErrors.labels(error_type="http").inc()
                self.metrics.counter("error.http").increment()

                # we should crash if it's our fault
                response = getattr(exc, "response", None)
                if response is not None and response.status_code < 500:
                    logger.exception("HTTP Request failed. Error: %s", response.text)
                    if response.status_code != 422:
                        # Do not exit on validation errors
                        raise
                else:
                    logger.exception("HTTP Request failed.")
            except OSError:
                publishErrors.labels(error_type="io").inc()
                self.metrics.counter("error.io").increment()
                logger.exception("HTTP Request failed")
            else:
                publishCountTotal.inc(payload.item_count)
                self.metrics.counter("sent").increment(payload.item_count)
                return

        raise MaxRetriesError("could not sent batch")


SERIALIZER_BY_VERSION = {"2": V2Batch, "2j": V2JBatch}


def publish_events() -> None:
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "config_file", type=argparse.FileType("r"), help="path to a configuration file"
    )
    arg_parser.add_argument(
        "--queue-name",
        default="main",
        help="name of event queue / publisher config (default: main)",
    )
    arg_parser.add_argument(
        "--debug", default=False, action="store_true", help="enable debug logging"
    )
    args = arg_parser.parse_args()

    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.WARNING
    logging.basicConfig(level=level)

    config_parser = configparser.RawConfigParser(interpolation=EnvironmentInterpolation())
    config_parser.read_file(args.config_file)
    raw_config = dict(config_parser.items("event-publisher:" + args.queue_name))
    cfg = config.parse_config(
        raw_config,
        {
            "collector": {
                "hostname": config.String,
                "version": config.Optional(config.String, default="2"),
                "scheme": config.Optional(config.String, default="https"),
            },
            "key": {"name": config.String, "secret": config.Base64},
            "max_queue_size": config.Optional(config.Integer, MAX_QUEUE_SIZE),
        },
    )

    metrics_client = metrics_client_from_config(raw_config)

    event_queue = MessageQueue(
        "/events-" + args.queue_name,
        max_messages=cfg.max_queue_size,
        max_message_size=MAX_EVENT_SIZE,
    )

    # pylint: disable=maybe-no-member
    serializer = SERIALIZER_BY_VERSION[cfg.collector.version]()
    batcher = TimeLimitedBatch(serializer, MAX_BATCH_AGE)
    publisher = BatchPublisher(metrics_client, cfg)

    while True:
        # allow other routines to execute (specifically handling requests to /metrics)
        gevent.sleep(0)
        message: Optional[bytes]

        try:
            message = event_queue.get(timeout=0.2)
        except TimedOutError:
            message = None

        try:
            batcher.add(message)
            continue
        except BatchFull:
            pass

        serialized = batcher.serialize()
        try:
            publisher.publish(serialized)
        except Exception:
            logger.exception("Events publishing failed.")
        batcher.reset()
        batcher.add(message)


if __name__ == "__main__":
    start_prometheus_exporter_for_sidecar()
    publish_events()



import abc
import datetime
import logging
import os
import queue
import signal
import socket
import uuid

from threading import Thread
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import TYPE_CHECKING

from gevent.pywsgi import LoggingLogAdapter
from gevent.pywsgi import WSGIServer
from gevent.server import StreamServer

import baseplate.lib.config

from baseplate.lib.retry import RetryPolicy
from baseplate.observers.timeout import ServerTimeout
from baseplate.server import runtime_monitor


logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from wsgiref.types import StartResponse  # pylint: disable=import-error,no-name-in-module

WSGIEnvironment = Dict[str, Any]
HealthcheckCallback = Callable[[WSGIEnvironment], bool]


class HealthcheckApp:
    def __init__(self, callback: Optional[HealthcheckCallback] = None) -> None:
        self.callback = callback

    def __call__(self, environ: WSGIEnvironment, start_response: "StartResponse") -> List[bytes]:
        ok = True
        if self.callback:
            ok = self.callback(environ)

        if ok:
            start_response("200 OK", [("Content-Type", "application/json")])
            return [b'{"fun message": "all good, boss"}']
        start_response("503 UNAVAILABLE", [("Content-Type", "text/plain")])
        return [b"he's dead jim"]


def make_simple_healthchecker(
    listener: socket.socket, callback: Optional[HealthcheckCallback] = None
) -> WSGIServer:
    return WSGIServer(
        listener=listener,
        application=HealthcheckApp(callback),
        log=LoggingLogAdapter(logger, level=logging.DEBUG),
    )


class PumpWorker(abc.ABC):
    """Reads messages off of a message queue and puts them into a queue.Queue for handling by a MessageHandler.

    The QueueConsumerServer will run a single PumpWorker in its own thread.
    """

    @abc.abstractmethod
    def run(self) -> None:
        """Run the worker.

        This is the method called by the the PumpWorker Thread within the
        QueueConsumerServer and is ultimately responsible for getting messages
        off of the message queue and putting them on the work queue.
        """

    @abc.abstractmethod
    def stop(self) -> None:
        """Signal the PumpWorker that it should stop receiving new messages from its message queue."""


class MessageHandler(abc.ABC):
    """Processes the messages supplied by the PumpWorker.

    A MessageHandler is expected to have a Baseplate object that it can use to
    generate a RequestContext and a ServerSpan before processing each message.

    The QueueConsumerServer will run multiple MessageHandlers, each in their own
    Thread based on the max_concurrency value it is given.  Each MessageHandler
    will be wrapped in a QueueConsumer object that will handle reading new
    messages off of the work queue and passing them to the MessageHandler as well
    as providing the methods that start and stop the work done by the Threads.
    """

    @abc.abstractmethod
    def handle(self, message: Any) -> None:
        """Handle a message supplied by a PumpWorker.

        This method is what the server calls when it reads a new message off of
        the internal work queue that is fed by the PumpWorker.  This is both the
        method where you should handle processing the message and setting up
        the RequestContext and ServerSpan for Baseplate. We recommend that that
        you only implement boilerplate, such as Baseplate setup and standard
        ack/requeue behavior for the message, in this method.  The actual business
        logic around handling a message should be given to the MessageHandler
        when it is constructed rather than defined in here.
        """


class QueueConsumerFactory(abc.ABC):
    """Factory for building all of the objects needed to run a QueueConsumerServer.

    This is the type of object that you build in your `service.make_queue_consumer`
    function to be passed to `make_server`.
    """

    @abc.abstractmethod
    def build_pump_worker(self, work_queue: queue.Queue) -> PumpWorker:
        """Build an object implementing the PumpWorker interface.

        `work_queue` is the Queue that will be shared between the PumpWorker and
        and the QueueConsumer to pass messages to the MessageHandler, be sure
        that your PumpWorker uses `work_queue`.
        """

    @abc.abstractmethod
    def build_message_handler(self) -> MessageHandler:
        """Build an object implementing the MessageHandler interface."""

    @abc.abstractmethod
    def build_health_checker(self, listener: socket.socket) -> StreamServer:
        """Build an HTTP server to service health checks."""


class QueueConsumer:
    """Wrapper around a MessageHandler object that interfaces with the work_queue and starts/stops the handle loop.

    This object is used by the QueueConsumerServer to wrap a MessageHandler object
    before creating a worker Thread.  This allows the MessageHandler to focus soley
    on handling a single message while the QueueConsumer pulls messages from the
    work_queue for the MessageHandler and handles starting/stopping based on the
    commands from the server.
    """

    def __init__(self, work_queue: queue.Queue, message_handler: MessageHandler):
        self.id = uuid.uuid4()
        self.work_queue = work_queue
        self.message_handler = message_handler
        self.started = False
        self.stopped = False
        self._queue_timeout = 5

    def stop(self) -> None:
        """Signal the QueueConsumer to stop processing."""
        assert self.started
        assert not self.stopped
        self.stopped = True

    def run(self) -> None:
        """Run the queue consumer until stopped or hit an unhandled Exception."""
        assert not self.started
        assert not self.stopped
        logger.debug("Consumer <%s> starting.", self.id)
        self.started = True
        while not self.stopped:
            try:
                # We set a timeout so we can periodically check if we should
                # stop, this way we will actually return if we have recieved a
                # `stop()` call rather than hanging in a `get` waiting for a
                # new message.  If we did not do this, we would not be able to
                # wait for all of our workers to finish before stopping the
                # server.
                message = self.work_queue.get(timeout=self._queue_timeout)
            except queue.Empty:
                pass
            else:
                # Ensure that if self.message_handler.handle throws a `queue.Empty`
                # error, that bubbles up and is not treated as though `self.work_queue`
                # is empty
                self.message_handler.handle(message)
        logger.debug("Consumer <%s> stopping.", self.id)


class QueueConsumerServer:
    """Server for running long-lived queue consumers."""

    def __init__(
        self,
        pump: PumpWorker,
        handlers: Sequence[QueueConsumer],
        healthcheck_server: StreamServer,
        stop_timeout: datetime.timedelta,
    ):
        self.pump = pump
        self.handlers = handlers
        self.healthcheck_server = healthcheck_server
        self.stop_timeout = stop_timeout

        def watcher(fn: Callable) -> Callable:
            """Terminates the server (gracefully) if `fn` raises an Exception.

            Used to monitor the pump and handler threads for Exceptions so we can
            shut down the server if one of them exits unexpectedly.
            """

            def _run_and_terminate(*a: Any, **kw: Any) -> Any:
                try:
                    return fn(*a, **kw)
                except (Exception, ServerTimeout):
                    logger.exception("Unhandled error in pump or handler thread, terminating.")
                    self._terminate()
                return None

            return _run_and_terminate

        self.pump_thread = Thread(target=watcher(self.pump.run), daemon=True)
        self.threads = [Thread(target=watcher(handler.run)) for handler in self.handlers]
        self.started = False
        self.stopped = False

    @classmethod
    def new(
        cls,
        max_concurrency: int,
        consumer_factory: QueueConsumerFactory,
        listener: socket.socket,
        stop_timeout: datetime.timedelta,
    ) -> "QueueConsumerServer":
        """Build a new QueueConsumerServer."""
        # We want to give some headroom on the queue so our handlers can grab
        # a new message right after they finish so we keep an extra
        # max_concurrency / 2 messages in the queue.
        maxsize = max_concurrency + max_concurrency // 2
        work_queue: queue.Queue = queue.Queue(maxsize=maxsize)
        handlers = [
            QueueConsumer(
                work_queue=work_queue, message_handler=consumer_factory.build_message_handler()
            )
            for _ in range(max_concurrency)
        ]
        return cls(
            pump=consumer_factory.build_pump_worker(work_queue),
            handlers=handlers,
            healthcheck_server=consumer_factory.build_health_checker(listener),
            stop_timeout=stop_timeout,
        )

    def _terminate(self) -> None:
        """Send a SIGTERM signal to ourselves so baseplate can call `stop`."""
        assert self.started
        if not self.stopped:
            os.kill(os.getpid(), signal.SIGTERM)

    def start(self) -> None:
        """Start the server.

        Starts all of the worker threads and the healthcheck server (if it exists).

        Should only be called once and should not be called after the server is
        stopped, will raise an AssertionError in either of those cases.
        """
        assert not self.started
        assert not self.stopped
        logger.debug("Starting server.")
        self.started = True
        logger.debug("Starting pump thread.")
        self.pump_thread.start()
        logger.debug("Starting message handler threads.")
        for thread in self.threads:
            thread.start()
        logger.debug("Starting healthcheck server.")
        self.healthcheck_server.start()
        logger.debug("Server started.")

    def stop(self) -> None:
        """Start the server.

        Stop all of the worker threads and the healthcheck server (if it exists).
        Waits for the handler threads to drain before returning but does not wait
        for the pump or watcher threads to finish.

        Should only be called once and should not be before the server is
        started, will raise an AssertionError in either of those cases.
        """
        assert self.started
        assert not self.stopped
        logger.debug("Stopping server.")
        self.stopped = True
        # Stop the pump first so we stop consuming messages from the message
        # queue
        logger.debug("Stopping pump thread.")
        self.pump.stop()
        # It's important to call `handler.stop()` before calling `join` on the
        # handler threads, otherwise we'll be waiting for threads that have not
        # been instructed to stop.
        logger.debug("Stopping message handler threads.")
        for handler in self.handlers:
            handler.stop()
        retry_policy = RetryPolicy.new(budget=self.stop_timeout.total_seconds())
        logger.debug("Waiting for message handler threads to drain.")
        for time_remaining, thread in zip(retry_policy, self.threads):
            thread.join(timeout=time_remaining)
        # Stop the healthcheck server last
        logger.debug("Stopping healthcheck server.")
        self.healthcheck_server.stop()
        logger.debug("Server stopped.")


def make_server(
    server_config: Dict[str, str], listener: socket.socket, app: QueueConsumerFactory
) -> QueueConsumerServer:
    """Make a queue consumer server for long running queue consumer apps.

    If you require that you process messages in-order, your handler is heavily CPU
    bound, or you don't do any IO when handling a message you should restrict
    max_concurrency to 1.
    """
    cfg = baseplate.lib.config.parse_config(
        server_config,
        {
            "max_concurrency": baseplate.lib.config.Integer,
            "stop_timeout": baseplate.config.Optional(
                baseplate.config.Timespan, default=datetime.timedelta(seconds=30)
            ),
        },
    )

    runtime_monitor.start(server_config, app, pool=None)

    return QueueConsumerServer.new(
        consumer_factory=app,
        max_concurrency=cfg.max_concurrency,
        listener=listener,
        stop_timeout=cfg.stop_timeout,
    )



from __future__ import annotations

import logging

from types import TracebackType
from typing import Any
from typing import List
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import Union

import sentry_sdk

from baseplate import _ExcInfo
from baseplate import BaseplateObserver
from baseplate import RequestContext
from baseplate import ServerSpanObserver
from baseplate import Span
from baseplate.lib import config
from baseplate.observers.timeout import ServerTimeout

if TYPE_CHECKING:
    from gevent.hub import Hub as GeventHub


ALWAYS_IGNORE_ERRORS = (
    "baseplate.observers.timeout.ServerTimeout",
    "ConnectionError",
    "ConnectionRefusedError",
    "ConnectionResetError",
    "pymemcache.exceptions.MemcacheServerError",
    "requests.exceptions.HTTPError",
    "thrift.Thrift.TApplicationException",
    "thrift.Thrift.TProtocolException",
    "thrift.Thrift.TTransportException",
)


def init_sentry_client_from_config(raw_config: config.RawConfig, **kwargs: Any) -> None:
    """Configure the Sentry client.

    This expects one configuration option and can take many optional ones:

    ``sentry.dsn``
        The DSN provided by Sentry. If blank, the reporter will discard events.
    ``sentry.environment`` (optional)
        The environment your application is running in.
    ``sentry.sample_rate`` (optional)
        Percentage of errors to report. (e.g. "37%")
    ``sentry.ignore_errors`` (optional)
        A comma-delimited list of exception names, unqualified (e.g.
        ServerTimeout) or fully qualified (e.g.
        baseplate.observers.timeout.ServerTimeout) to not notify sentry about.
        Note: a minimal list of common exceptions is hard-coded in Baseplate,
        this option only extends that list.

    Example usage::

        init_sentry_client_from_config(app_config)

    :param raw_config: The application configuration which should have
        settings for the error reporter.

    """
    cfg = config.parse_config(
        raw_config,
        {
            "sentry": {
                "dsn": config.Optional(config.String, default=None),
                "environment": config.Optional(config.String, default=None),
                "sample_rate": config.Optional(config.Percent, default=1),
                "ignore_errors": config.Optional(config.TupleOf(config.String), default=()),
            }
        },
    )

    if cfg.sentry.dsn:
        kwargs.setdefault("dsn", cfg.sentry.dsn)

    if cfg.sentry.environment:
        kwargs.setdefault("environment", cfg.sentry.environment)

    kwargs.setdefault("sample_rate", cfg.sentry.sample_rate)

    ignore_errors: List[Union[type, str]] = []
    ignore_errors.extend(ALWAYS_IGNORE_ERRORS)
    ignore_errors.extend(cfg.sentry.ignore_errors)
    kwargs.setdefault("ignore_errors", ignore_errors)

    kwargs.setdefault("with_locals", False)

    client = sentry_sdk.Client(**kwargs)
    sentry_sdk.Hub.current.bind_client(client)


class SentryBaseplateObserver(BaseplateObserver):
    """Error reporting observer.

    This observer reports unexpected exceptions to Sentry.

    """

    def on_server_span_created(self, context: RequestContext, server_span: Span) -> None:
        sentry_hub = sentry_sdk.Hub.current
        observer = _SentryServerSpanObserver(sentry_hub, server_span)
        server_span.register(observer)
        context.sentry = sentry_hub


class _SentryServerSpanObserver(ServerSpanObserver):
    def __init__(self, sentry_hub: sentry_sdk.Hub, server_span: Span):
        self.sentry_hub = sentry_hub
        self.scope_manager = self.sentry_hub.push_scope()
        self.scope = self.scope_manager.__enter__()
        self.server_span = server_span

    def on_start(self) -> None:
        self.scope.set_tag("trace_id", self.server_span.trace_id)

    def on_set_tag(self, key: str, value: Any) -> None:
        self.scope.set_tag(key, value)

    def on_log(self, name: str, payload: Any) -> None:
        self.sentry_hub.add_breadcrumb({"category": name, "message": str(payload)})

    def on_finish(self, exc_info: Optional[_ExcInfo] = None) -> None:
        if exc_info is not None:
            self.sentry_hub.capture_exception(error=exc_info)
        self.scope_manager.__exit__(None, None, None)


class _SentryUnhandledErrorReporter:
    """Hook into the Gevent hub and report errors outside request context."""

    @classmethod
    def install(cls) -> None:
        from gevent import get_hub

        gevent_hub = get_hub()
        gevent_hub.print_exception = cls(gevent_hub)

    @classmethod
    def uninstall(cls) -> None:
        from gevent import get_hub

        gevent_hub = get_hub()
        assert isinstance(gevent_hub.print_exception, cls)
        gevent_hub.print_exception = gevent_hub.print_exception.original_print_exception

    def __init__(self, hub: GeventHub):
        self.original_print_exception = getattr(hub, "print_exception")
        self.logger = logging.getLogger(__name__)

    def __call__(
        self,
        context: Any,
        exc_type: Optional[Type[BaseException]],
        value: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        sentry_sdk.capture_exception((exc_type, value, tb))

        if value and isinstance(value, ServerTimeout):
            self.logger.warning(
                "Server timed out processing for %r after %0.2f seconds",
                value.span_name,
                value.timeout_seconds,
                exc_info=(exc_type, value, tb) if value.debug else None,  # type: ignore
            )
        else:
            self.original_print_exception(context, exc_type, value, tb)



import threading

from baseplate import BaseplateObserver
from baseplate import RequestContext
from baseplate import Span


class LoggingBaseplateObserver(BaseplateObserver):
    """Logging observer.

    This observer adds request context to the thread-local state so that the
    log formatters can give more informative logs. Currently, this just sets
    the thread name to the current request's trace ID.

    """

    def on_server_span_created(self, context: RequestContext, server_span: Span) -> None:
        threading.current_thread().name = str(server_span.trace_id)



"""The baseplate server.

This command serves your application from the given configuration file.

"""
import argparse
import code
import configparser
import enum
import gc
import importlib
import inspect
import logging.config
import os
import signal
import socket
import sys
import syslog
import threading
import time
import traceback
import warnings

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from rlcompleter import Completer
from types import FrameType
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Mapping
from typing import MutableMapping
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import TextIO
from typing import Tuple

from gevent.server import StreamServer

from baseplate import Baseplate
from baseplate.lib import warn_deprecated
from baseplate.lib.config import Endpoint
from baseplate.lib.config import EndpointConfiguration
from baseplate.lib.config import Optional as OptionalConfig
from baseplate.lib.config import parse_config
from baseplate.lib.config import Timespan
from baseplate.lib.log_formatter import CustomJsonFormatter
from baseplate.lib.prometheus_metrics import is_metrics_enabled
from baseplate.server import einhorn
from baseplate.server import reloader
from baseplate.server.net import bind_socket


logger = logging.getLogger(__name__)


class ServerLifecycle(Enum):
    RUNNING = enum.auto()
    SHUTTING_DOWN = enum.auto()


@dataclass
class ServerState:
    state: ServerLifecycle = ServerLifecycle.RUNNING

    @property
    def shutting_down(self) -> bool:
        warn_deprecated("SERVER_STATE.shutting_down is deprecated in favor of SERVER_STATE.state")
        return self.state == ServerLifecycle.SHUTTING_DOWN


SERVER_STATE = ServerState()


def parse_args(args: Sequence[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=sys.modules[__name__].__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--debug", action="store_true", default=False, help="enable extra-verbose debug logging"
    )
    parser.add_argument(
        "--reload",
        action="store_true",
        default=False,
        help="restart the server when code changes (development only)",
    )
    parser.add_argument(
        "--app-name",
        default="main",
        metavar="NAME",
        help="name of app to load from config_file (default: main)",
    )
    parser.add_argument(
        "--server-name",
        default="main",
        metavar="NAME",
        help="name of server to load from config_file (default: main)",
    )
    parser.add_argument(
        "--bind",
        type=Endpoint,
        default=Endpoint("localhost:9090"),
        metavar="ENDPOINT",
        help="endpoint to bind to (ignored if under Einhorn)",
    )
    parser.add_argument(
        "config_file", type=argparse.FileType("r"), help="path to a configuration file"
    )

    return parser.parse_args(args)


class EnvironmentInterpolation(configparser.Interpolation):
    def before_get(
        self,
        parser: MutableMapping[str, Mapping[str, str]],
        section: str,
        option: str,
        value: str,
        defaults: Mapping[str, str],
    ) -> str:
        return os.path.expandvars(value)


class Configuration(NamedTuple):
    filename: str
    server: Optional[Dict[str, str]]
    app: Dict[str, str]
    has_logging_options: bool
    shell: Optional[Dict[str, str]]


def read_config(config_file: TextIO, server_name: Optional[str], app_name: str) -> Configuration:
    # we use RawConfigParser to reduce surprise caused by interpolation and so
    # that config.Percent works more naturally (no escaping %).
    parser = configparser.RawConfigParser(interpolation=EnvironmentInterpolation())
    parser.read_file(config_file)

    filename = config_file.name
    server_config = dict(parser.items("server:" + server_name)) if server_name else None
    app_config = dict(parser.items("app:" + app_name))
    has_logging_config = parser.has_section("loggers")
    shell_config = None
    if parser.has_section("shell"):
        shell_config = dict(parser.items("shell"))
    elif parser.has_section("tshell"):
        shell_config = dict(parser.items("tshell"))

    return Configuration(filename, server_config, app_config, has_logging_config, shell_config)


def configure_logging(config: Configuration, debug: bool) -> None:
    logging.captureWarnings(capture=True)

    if debug:
        logging_level = logging.DEBUG
        warnings.simplefilter("always")  # enable DeprecationWarning etc.
    else:
        logging_level = logging.INFO

    formatter: logging.Formatter
    if not sys.stdin.isatty():
        formatter = CustomJsonFormatter(
            "%(levelname)s %(message)s %(funcName)s %(lineno)d %(module)s %(name)s %(pathname)s %(process)d %(processName)s %(thread)d %(threadName)s"
        )
    else:
        formatter = logging.Formatter("%(levelname)-8s %(message)s")

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    root_logger.setLevel(logging_level)
    root_logger.addHandler(handler)

    # add PID 1 stdout logging if we're containerized and not running under an init system
    if _is_containerized() and not _has_PID1_parent():
        file_handler = logging.FileHandler("/proc/1/fd/1", mode="w")
        file_handler.setFormatter(formatter)
        root_logger.addHandler(file_handler)

    if config.has_logging_options:
        logging.config.fileConfig(config.filename)


def make_listener(endpoint: EndpointConfiguration) -> socket.socket:
    try:
        return einhorn.get_socket()
    except (einhorn.NotEinhornWorker, IndexError):
        # we're not under einhorn or it didn't bind any sockets for us
        pass

    return bind_socket(endpoint)


def _load_factory(url: str, default_name: Optional[str] = None) -> Callable:
    """Load a factory function from a config file."""
    module_name, sep, func_name = url.partition(":")
    if not sep:
        if not default_name:
            raise ValueError("no name and no default specified")
        func_name = default_name
    module = importlib.import_module(module_name)
    factory = getattr(module, func_name)
    return factory


def make_server(
    server_config: Dict[str, str], listener: socket.socket, app: Callable
) -> StreamServer:
    server_url = server_config["factory"]
    factory = _load_factory(server_url, default_name="make_server")
    return factory(server_config, listener, app)


def make_app(app_config: Dict[str, str]) -> Callable:
    app_url = app_config["factory"]
    factory = _load_factory(app_url, default_name="make_app")
    return factory(app_config)


def register_signal_handlers() -> threading.Event:
    def _handle_debug_signal(_signo: int, frame: FrameType) -> None:
        if not frame:
            logger.warning("Received SIGUSR1, but no frame found.")
            return

        lines = traceback.format_stack(frame)
        logger.warning("Received SIGUSR1, dumping stack trace:")
        for line in lines:
            logger.warning(line.rstrip("\n"))

    signal.signal(signal.SIGUSR1, _handle_debug_signal)
    signal.siginterrupt(signal.SIGUSR1, False)

    shutdown_event = threading.Event()

    def _handle_shutdown_signal(_signo: int, _frame: FrameType) -> None:
        shutdown_event.set()

    # shutdown is signalled differently in different contexts:
    # - SIGINT  - Ctrl-C on the command line
    # - SIGTERM - Kubernetes
    # - SIGUSR2 - Einhorn
    for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGUSR2):
        signal.signal(sig, _handle_shutdown_signal)
        signal.siginterrupt(sig, False)
    return shutdown_event


def load_app_and_run_server() -> None:
    """Parse arguments, read configuration, and start the server."""
    sys.path.append(os.getcwd())

    shutdown_event = register_signal_handlers()

    args = parse_args(sys.argv[1:])
    with args.config_file:
        config = read_config(args.config_file, args.server_name, args.app_name)
    assert config.server

    configure_logging(config, args.debug)

    app = make_app(config.app)
    listener = make_listener(args.bind)
    server = make_server(config.server, listener, app)

    if einhorn.is_worker():
        einhorn.ack_startup()

    if is_metrics_enabled(config.app):
        from baseplate.server.prometheus import start_prometheus_exporter

        start_prometheus_exporter()
    else:
        logger.info("Metrics are not configured, Prometheus metrics will not be exported.")

    if args.reload:
        reloader.start_reload_watcher(extra_files=[args.config_file.name])

    # clean up leftovers from initialization before we get into requests
    gc.collect()

    logger.info("Listening on %s", listener.getsockname())
    server.start()
    try:
        shutdown_event.wait()

        SERVER_STATE.state = ServerLifecycle.SHUTTING_DOWN

        cfg = parse_config(config.server, {"drain_time": OptionalConfig(Timespan)})
        if cfg.drain_time:
            logger.debug("Draining inbound requests...")
            time.sleep(cfg.drain_time.total_seconds())
    finally:
        logger.debug("Gracefully shutting down...")
        server.stop()
        logger.info("Exiting")


def load_and_run_script() -> None:
    """Launch a script with an entrypoint similar to a server."""
    sys.path.append(os.getcwd())

    args, extra_args = _parse_baseplate_script_args()
    with args.config_file:
        config = read_config(args.config_file, server_name=None, app_name=args.app_name)
    configure_logging(config, args.debug)

    entrypoint = _load_factory(args.entrypoint)

    if _fn_accepts_additional_args(entrypoint, extra_args):
        entrypoint(config.app, extra_args)
    else:
        entrypoint(config.app)


def _parse_baseplate_script_args() -> Tuple[argparse.Namespace, List[str]]:
    parser = argparse.ArgumentParser(
        description="Run a function with app configuration loaded.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--debug", action="store_true", default=False, help="enable extra-verbose debug logging"
    )
    parser.add_argument(
        "--app-name",
        default="main",
        metavar="NAME",
        help="name of app to load from config_file (default: main)",
    )
    parser.add_argument(
        "config_file", type=argparse.FileType("r"), help="path to a configuration file"
    )
    parser.add_argument("entrypoint", type=str, help="function to call, e.g. module.path:fn_name")
    return parser.parse_known_args(sys.argv[1:])


def _fn_accepts_additional_args(script_fn: Callable[..., Any], fn_args: Sequence[str]) -> bool:
    additional_args_provided = len(fn_args) > 0
    signature = inspect.signature(script_fn)

    positional_arg_count = 0
    allows_var_args = False
    for param in signature.parameters.values():
        if param.kind in {param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD}:
            positional_arg_count += 1
        elif param.kind == param.VAR_POSITIONAL:
            allows_var_args = True

    allows_additional_args = allows_var_args or positional_arg_count > 1

    if positional_arg_count < 1 and not allows_var_args:
        raise ValueError("script function accepts too few positional arguments")
    if positional_arg_count > 2:
        raise ValueError("script function accepts too many positional arguments")
    if additional_args_provided and not allows_additional_args:
        raise ValueError(
            "script function does not accept additional arguments, "
            "but additional arguments were provided"
        )

    return allows_additional_args


def load_and_run_shell() -> None:
    """Launch a shell for a thrift service."""
    sys.path.append(os.getcwd())

    parser = argparse.ArgumentParser(
        description="Open a shell for a Thrift service with app configuration loaded.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--debug", action="store_true", default=False, help="enable extra-verbose debug logging"
    )
    parser.add_argument(
        "--app-name",
        default="main",
        metavar="NAME",
        help="name of app to load from config_file (default: main)",
    )
    parser.add_argument(
        "config_file", type=argparse.FileType("r"), help="path to a configuration file"
    )

    args = parser.parse_args(sys.argv[1:])
    with args.config_file:
        config = read_config(args.config_file, server_name=None, app_name=args.app_name)
    logging.basicConfig(level=logging.INFO)

    env: Dict[str, Any] = {}
    env_banner = {
        "app": "This project's app instance",
        "context": "The context for this shell instance's span",
    }

    app = make_app(config.app)
    env["app"] = app

    baseplate: Baseplate = app.baseplate  # type: ignore
    context = baseplate.make_context_object()
    span = baseplate.make_server_span(context, "shell")
    env["context"] = span.context

    if config.shell and "setup" in config.shell:
        setup = _load_factory(config.shell["setup"])
        setup(env, env_banner)

    configure_logging(config, args.debug)

    # generate banner text
    banner = "Available Objects:\n"
    for var in sorted(env_banner.keys()):
        banner += f"\n  {var:<12} {env_banner[var]}"

    console_logpath = _get_shell_log_path()

    try:
        # try to use IPython if possible
        from IPython import start_ipython

        try:
            # IPython 5.x+
            from traitlets.config.loader import Config
        except ImportError:
            # IPython 4 and below
            from IPython import Config

        ipython_config = Config()
        ipython_config.InteractiveShellApp.exec_lines = [
            # monkeypatch IPython's log-write() to enable formatted input logging, copying original code:
            # https://github.com/ipython/ipython/blob/a54bf00feb5182fa821bd5457897b3b30a313436/IPython/core/logger.py#L187-L201
            f"""
            ip = get_ipython()
            from functools import partial
            def log_write(self, data, kind="input", message_id="IEXC"):
                import datetime, os
                if self.log_active and data:
                    write = self.logfile.write
                    if kind=='input':
                        # Generate an RFC 5424 compliant syslog format
                        write(f'<13>1 {{datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ")}} {{os.uname().nodename}} baseplate-shell {{os.getpid()}} {{message_id}} - {{data}}')
                    elif kind=='output' and self.log_output:
                        odata = u'\\n'.join([u'#[Out]# %s' % s
                                        for s in data.splitlines()])
                        write(u'%s\\n' % odata)
                    self.logfile.flush()
            ip.logger.logstop = None
            ip.logger.log_write = partial(log_write, ip.logger)
            ip.magic('logstart {console_logpath} append')
            ip.logger.log_write(data="Start IPython logging\\n", message_id="ISTR")
            """
        ]
        ipython_config.TerminalInteractiveShell.banner2 = banner
        ipython_config.LoggingMagics.quiet = True
        start_ipython(argv=[], user_ns=env, config=ipython_config)
        raise SystemExit
    except ImportError:
        pass

    newbanner = f"Baseplate Interactive Shell\nPython {sys.version}\n\n"
    banner = newbanner + banner

    try:
        import readline

        readline.set_completer(Completer(env).complete)
        readline.parse_and_bind("tab: complete")

    except ImportError:
        pass

    shell = LoggedInteractiveConsole(_locals=env, logpath=console_logpath)
    shell.interact(banner)


def _get_shell_log_path() -> str:
    """Determine where to write shell audit logs."""
    if _is_containerized():
        # write to PID 1 stdout for log aggregation
        return "/proc/1/fd/1"
    # otherwise write to a local file
    return "/var/log/baseplate-shell.log"


def _is_containerized() -> bool:
    """Determine if we're running in a container based on cgroup awareness for various container runtimes."""
    if os.path.exists("/.dockerenv"):
        return True

    try:
        with open("/proc/self/cgroup", encoding="UTF-8") as my_cgroups_file:
            my_cgroups = my_cgroups_file.read()

            for hint in ["kubepods", "docker", "containerd"]:
                if hint in my_cgroups:
                    return True
    except OSError:
        pass

    return False


def _has_PID1_parent() -> bool:
    """Determine parent PIDs up the tree until PID 1 or 0 is reached, do this natively"""
    parent_pid = os.getppid()
    while parent_pid > 1:
        with open(f"/proc/{parent_pid}/status", encoding="UTF-8") as proc_status:
            for line in proc_status.readlines():
                if line.startswith("PPid:"):
                    parent_pid = int(line.replace("PPid:", ""))
                    break
    return bool(parent_pid)


class LoggedInteractiveConsole(code.InteractiveConsole):
    def __init__(self, _locals: Dict[str, Any], logpath: str) -> None:
        code.InteractiveConsole.__init__(self, _locals)
        self.output_file = logpath
        self.pid = os.getpid()
        self.pri = syslog.LOG_USER | syslog.LOG_NOTICE
        self.hostname = os.uname().nodename
        self.log_event(message="Start InteractiveConsole logging", message_id="CSTR")

    def raw_input(self, prompt: Optional[str] = "") -> str:
        data = input(prompt)
        self.log_event(message=data, message_id="CEXC")
        return data

    def log_event(
        self, message: str, message_id: Optional[str] = "-", structured: Optional[str] = "-"
    ) -> None:
        """Generate an RFC 5424 compliant syslog format."""
        timestamp = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        prompt = f"<{self.pri}>1 {timestamp} {self.hostname} baseplate-shell {self.pid} {message_id} {structured} {message}"
        with open(self.output_file, "w", encoding="UTF-8") as f:
            print(prompt, file=f)
            f.flush()



import json
import unittest

from unittest import mock

from baseplate import ServerSpan
from baseplate import Span
from baseplate.lib.config import Endpoint
from baseplate.observers.tracing import ANNOTATIONS
from baseplate.observers.tracing import LoggingRecorder
from baseplate.observers.tracing import make_client
from baseplate.observers.tracing import NullRecorder
from baseplate.observers.tracing import RemoteRecorder
from baseplate.observers.tracing import TraceBaseplateObserver
from baseplate.observers.tracing import TraceLocalSpanObserver
from baseplate.observers.tracing import TraceServerSpanObserver
from baseplate.observers.tracing import TraceSpanObserver


class TraceTestBase(unittest.TestCase):
    def setUp(self):
        thread_patch = mock.patch("threading.Thread", autospec=True)
        thread_patch.start()
        self.addCleanup(thread_patch.stop)


class TraceObserverTests(TraceTestBase):
    def setUp(self):
        super().setUp()
        self.mock_context = mock.Mock()

    def test_null_recorder_setup(self):
        client = make_client("test-service", log_if_unconfigured=False)
        baseplate_observer = TraceBaseplateObserver(client)
        self.assertEqual(type(baseplate_observer.recorder), NullRecorder)

    def test_logging_recorder_setup(self):
        client = make_client("test-service")
        baseplate_observer = TraceBaseplateObserver(client)
        self.assertEqual(type(baseplate_observer.recorder), LoggingRecorder)

    def test_sets_hostname(self):
        client = make_client("test-service")
        baseplate_observer = TraceBaseplateObserver(client)
        self.assertIsNotNone(baseplate_observer.hostname)

    def test_remote_recorder_setup(self):
        client = make_client("test-service", tracing_endpoint=Endpoint("test:1111"))
        baseplate_observer = TraceBaseplateObserver(client)
        self.assertTrue(isinstance(baseplate_observer.recorder, RemoteRecorder))

    def test_register_server_span_observer(self):
        client = make_client("test-service")
        baseplate_observer = TraceBaseplateObserver(client)
        context_mock = mock.Mock()
        span = ServerSpan(
            "test-id", "test-parent-id", "test-span-id", True, 0, "test", self.mock_context
        )
        baseplate_observer.on_server_span_created(context_mock, span)
        self.assertEqual(len(span.observers), 1)
        self.assertEqual(type(span.observers[0]), TraceServerSpanObserver)

    def test_force_sampling(self):
        span_with_debug_flag = Span(
            "test-id", "test-parent", "test-span-id", True, 1, "test", self.mock_context
        )
        span_without_debug_flag = Span(
            "test-id", "test-parent", "test-span-id", True, 0, "test", self.mock_context
        )
        self.assertTrue(TraceBaseplateObserver.force_sampling(span_with_debug_flag))
        self.assertFalse(TraceBaseplateObserver.force_sampling(span_without_debug_flag))

    def test_should_sample_utilizes_sampled_setting(self):
        client = make_client("test-service", sample_rate=0)
        baseplate_observer = TraceBaseplateObserver(client)
        span_with_sampled_flag = Span(
            "test-id", "test-parent", "test-span-id", True, 0, "test", self.mock_context
        )
        self.assertTrue(baseplate_observer.should_sample(span_with_sampled_flag))

    def test_should_sample_utilizes_force_sampling(self):
        client = make_client("test-service", sample_rate=0)
        baseplate_observer = TraceBaseplateObserver(client)
        span_with_forced = Span(
            "test-id", "test-parent", "test-span-id", False, 1, "test", self.mock_context
        )
        span_without_forced = Span(
            "test-id", "test-parent", "test-span-id", False, 0, "test", self.mock_context
        )
        self.assertTrue(baseplate_observer.should_sample(span_with_forced))
        self.assertFalse(baseplate_observer.should_sample(span_without_forced))

    def test_should_sample_utilizes_sample_rate(self):
        client = make_client("test-service", sample_rate=1)
        baseplate_observer = TraceBaseplateObserver(client)
        span = Span("test-id", "test-parent", "test-span-id", None, 0, "test", self.mock_context)
        self.assertTrue(baseplate_observer.should_sample(span))
        baseplate_observer.sample_rate = 0
        self.assertFalse(baseplate_observer.should_sample(span))

    def test_no_tracing_without_sampling(self):
        client = make_client("test-service", sample_rate=0)
        baseplate_observer = TraceBaseplateObserver(client)
        context_mock = mock.Mock()
        span = ServerSpan(
            "test-id", "test-parent-id", "test-span-id", False, 0, "test", self.mock_context
        )
        baseplate_observer.on_server_span_created(context_mock, span)
        self.assertEqual(len(span.observers), 0)


class TraceSpanObserverTests(TraceTestBase):
    def setUp(self):
        super().setUp()
        self.recorder = NullRecorder()
        self.mock_context = mock.Mock()

        self.span = Span(
            "test-id", "test-parent-id", "test-span-id", None, 0, "test", self.mock_context
        )

        self.debug_span = Span(
            "test-id", "test-parent-id", "test-span-id", None, 1, "test", self.mock_context
        )

        self.test_span_observer = TraceSpanObserver(
            "test-service", "test-hostname", self.span, self.recorder
        )

        self.test_debug_span_observer = TraceSpanObserver(
            "test-service", "test-hostname", self.debug_span, self.recorder
        )

    def test_component_set_on_initialization(self):
        component_set = False
        for annotation in self.test_span_observer.binary_annotations:
            if annotation["key"] == ANNOTATIONS["COMPONENT"] and annotation["value"] == "baseplate":
                component_set = True
                break
        self.assertTrue(component_set)

    def test_debug_span_tag_set_on_initialization(self):

        for annotation in self.test_debug_span_observer.binary_annotations:
            if annotation["key"] == ANNOTATIONS["DEBUG"]:
                self.assertTrue(annotation["value"])
                break

    def test_serialize_uses_span_info(self):
        serialized_span = self.test_span_observer._serialize()
        self.assertEqual(serialized_span["traceId"], self.span.trace_id)
        self.assertEqual(serialized_span["name"], self.span.name)
        self.assertEqual(serialized_span["id"], self.span.id)

    def test_serialize_adds_cs(self):
        serialized_span = self.test_span_observer._serialize()
        cs_in_annotation = False
        for annotation in serialized_span["annotations"]:
            if annotation["value"] == "cs":
                cs_in_annotation = True
        self.assertTrue(cs_in_annotation)

    def test_serialize_adds_cr(self):
        serialized_span = self.test_span_observer._serialize()
        cr_in_annotation = False
        for annotation in serialized_span["annotations"]:
            if annotation["value"] == "cr":
                cr_in_annotation = True
        self.assertTrue(cr_in_annotation)

    def test_serialize_adds_binary_annotations(self):
        # Reset in case initialization annotations exist.
        self.test_span_observer.binary_annotations = []

        self.test_span_observer.on_set_tag("test-key", "test-value")
        serialized_span = self.test_span_observer._serialize()
        self.assertEqual(len(serialized_span["binaryAnnotations"]), 1)
        annotation = serialized_span["binaryAnnotations"][0]
        self.assertEqual(annotation["key"], "test-key")
        self.assertEqual(annotation["value"], "test-value")

    def test_on_start_sets_start_timestamp(self):
        # on-start should set start time
        self.assertIsNone(self.test_span_observer.start)
        self.test_span_observer.on_start()
        self.assertIsNotNone(self.test_span_observer.start)

    def test_on_finish_sets_end_timestamp_and_duration(self):
        self.assertIsNone(self.test_span_observer.end)
        self.test_span_observer.on_start()
        self.test_span_observer.on_finish(None)
        self.assertIsNotNone(self.test_span_observer.end)

    def test_on_finish_records(self):
        self.assertIsNone(self.test_span_observer.end)
        self.test_span_observer.on_start()
        self.test_span_observer.on_finish(None)
        self.assertIsNotNone(self.test_span_observer.end)

    def test_on_finish_sets_debug_annotation(self):
        self.assertIsNone(self.test_debug_span_observer.end)
        self.test_debug_span_observer.on_start()
        self.test_debug_span_observer.on_finish(None)
        debug_annotation = None
        for annotation in self.test_debug_span_observer.binary_annotations:
            if annotation["key"] == ANNOTATIONS["DEBUG"]:
                debug_annotation = annotation
                break
        self.assertIsNotNone(debug_annotation)
        self.assertEqual(debug_annotation["value"], "true")

    def test_on_finish_sets_error_annotation(self):
        self.assertIsNone(self.test_span_observer.end)
        self.test_span_observer.on_start()
        self.test_span_observer.on_finish((ValueError, ValueError(), None))
        error_annotation = None
        for annotation in self.test_span_observer.binary_annotations:
            if annotation["key"] == ANNOTATIONS["ERROR"]:
                error_annotation = annotation
                break
        self.assertIsNotNone(error_annotation)
        self.assertEqual(error_annotation["value"], "true")

    def test_create_binary_annotation(self):
        annotation = self.test_span_observer._create_binary_annotation("test-key", "test-value")
        self.assertEqual(annotation["key"], "test-key")
        self.assertEqual(annotation["value"], "test-value")
        self.assertTrue(annotation["endpoint"])

    def test_create_binary_annotation_coerces_string(self):
        annotation = self.test_span_observer._create_binary_annotation("test-key", 1)
        self.assertEqual(annotation["key"], "test-key")
        self.assertEqual(annotation["value"], "1")
        self.assertTrue(annotation["endpoint"])

    def test_on_set_tag_adds_binary_annotation(self):
        self.test_span_observer.binary_annotations = []
        self.assertFalse(self.test_span_observer.binary_annotations)
        self.test_span_observer.on_set_tag("test-key", "test-value")
        annotation = self.test_span_observer.binary_annotations[0]
        self.assertEqual(annotation["key"], "test-key")
        self.assertEqual(annotation["value"], "test-value")

    def test_to_span_obj_sets_parent_id(self):
        span_obj = self.test_span_observer._to_span_obj([], [])
        self.assertEqual(span_obj["parentId"], self.span.parent_id)

    def test_to_span_obj_sets_default_parent_id(self):
        self.span.parent_id = None
        span_obj = self.test_span_observer._to_span_obj([], [])
        self.assertEqual(span_obj["parentId"], 0)

    def test_incr_tag_adds_binary_annotation(self):
        self.test_span_observer.binary_annotations = []
        self.test_span_observer.on_start()
        self.test_span_observer.on_incr_tag("test-key", 3)
        self.test_span_observer.on_incr_tag("test-key", 5)
        self.test_span_observer.on_finish(None)
        annotation = self.test_span_observer.binary_annotations[0]
        self.assertEqual(annotation["key"], "counter.test-key")
        self.assertEqual(annotation["value"], "8.0")


class TraceServerSpanObserverTests(TraceTestBase):
    def setUp(self):
        super().setUp()
        self.recorder = NullRecorder()
        self.mock_context = mock.Mock()
        self.span = ServerSpan(
            "test-id", "test-parent-id", "test-span-id", None, 0, "test", self.mock_context
        )
        self.test_server_span_observer = TraceServerSpanObserver(
            "test-service", "test-hostname", self.span, self.recorder
        )

    def test_server_span_observer_inherits_span_observer(self):
        # in case of future refactoring
        self.assertTrue(isinstance(self.test_server_span_observer, TraceSpanObserver))
        self.assertTrue(issubclass(TraceServerSpanObserver, TraceSpanObserver))

    def test_serialize_uses_span_info(self):
        serialized_span = self.test_server_span_observer._serialize()
        self.assertEqual(serialized_span["traceId"], self.span.trace_id)
        self.assertEqual(serialized_span["name"], self.span.name)
        self.assertEqual(serialized_span["id"], self.span.id)

    def test_serialize_adds_ss(self):
        serialized_span = self.test_server_span_observer._serialize()
        ss_in_annotation = False
        for annotation in serialized_span["annotations"]:
            if annotation["value"] == "ss":
                ss_in_annotation = True
        self.assertTrue(ss_in_annotation)

    def test_serialize_adds_sr(self):
        serialized_span = self.test_server_span_observer._serialize()
        sr_in_annotation = False
        for annotation in serialized_span["annotations"]:
            if annotation["value"] == "sr":
                sr_in_annotation = True
        self.assertTrue(sr_in_annotation)

    def test_on_child_span_created(self):
        child_span = Span(
            "child-id", "test-parent-id", "test-span-id", None, 0, "test-child", self.mock_context
        )
        self.test_server_span_observer.on_child_span_created(child_span)
        # Make sure new trace observer is added in the child span
        #  and the trace observer's span is that child span
        self.assertEqual(len(child_span.observers), 1)
        self.assertEqual(child_span.observers[0].span, child_span)

    def test_on_child_span_created_for_debug_span(self):
        child_span = Span(
            "child-id", "test-parent-id", "test-span-id", None, 1, "test-child", self.mock_context
        )
        self.test_server_span_observer.on_child_span_created(child_span)
        # Make sure new trace observer is added in the child span
        #  and the trace observer's span is that child span
        self.assertEqual(len(child_span.observers), 1)
        self.assertEqual(child_span.observers[0].span, child_span)


class TraceLocalSpanObserverTests(TraceTestBase):
    def setUp(self):
        super().setUp()
        self.recorder = NullRecorder()
        self.mock_context = mock.Mock()
        self.span = ServerSpan(
            "test-id", "test-parent-id", "test-span-id", None, 0, "test", self.mock_context
        )

    def test_init_local_component(self):
        local_trace_observer = TraceLocalSpanObserver(
            "test-service", "test-component", "test-host", self.span, self.recorder
        )
        self.assertEqual(local_trace_observer.component_name, "test-component")

    def test_local_component_structure(self):
        local_trace_observer = TraceLocalSpanObserver(
            "test-service", "test-component", "test-host", self.span, self.recorder
        )
        self.assertListEqual(
            local_trace_observer.binary_annotations,
            [
                {
                    "key": ANNOTATIONS["COMPONENT"],
                    "value": "baseplate",
                    "endpoint": {"ipv4": "test-host", "serviceName": "test-service"},
                },
                {
                    "key": ANNOTATIONS["LOCAL_COMPONENT"],
                    "value": "test-component",
                    "endpoint": {"ipv4": "test-host", "serviceName": "test-service"},
                },
            ],
        )

    def test_serialize(self):
        local_trace_observer = TraceLocalSpanObserver(
            "test-service", "test-component", "test-host", self.span, self.recorder
        )
        local_trace_observer.on_start()
        local_trace_observer.on_finish(None)
        serialized_span = local_trace_observer._serialize()
        self.assertIsNotNone(serialized_span["duration"])
        self.assertEqual(serialized_span["name"], self.span.name)
        annotations = serialized_span["binaryAnnotations"]
        for annotation in annotations:
            self.assertTrue("key" in annotation)
            self.assertTrue("value" in annotation)
            self.assertTrue("endpoint" in annotation)


class NullRecorderTests(TraceTestBase):
    def setUp(self):
        super().setUp()
        self.recorder = NullRecorder()
        self.mock_context = mock.Mock()

    def test_null_recorder_flush(self):
        span = Span("test-id", "test-parent-id", "test-span-id", None, 0, "test", self.mock_context)
        self.recorder.flush_func([span])


class RemoteRecorderTests(TraceTestBase):
    def setUp(self):
        super().setUp()
        self.endpoint = "test:1111"

    def test_init(self):
        recorder = RemoteRecorder(self.endpoint, 5)
        self.assertEqual(recorder.endpoint, "http://test:1111/api/v1/spans")

    def test_remote_recorder_flush(self):
        recorder = RemoteRecorder(self.endpoint, 5)
        serialized_span = {
            "traceId": "test-id",
            "name": "test-span",
            "id": "span-id",
            "timestamp": 0,
            "duration": 0,
            "annotations": [],
            "binaryAnnotations": [],
        }
        func_mock = mock.Mock()
        with mock.patch.object(recorder.session, "post", func_mock):
            recorder.flush_func([serialized_span])
            func_mock.assert_called_with(
                recorder.endpoint,
                data=json.dumps([serialized_span]).encode("utf8"),
                headers={"Content-Type": "application/json"},
                timeout=1,
            )



# Pylint documentation for writing a checker: http://pylint.pycqa.org/en/latest/how_tos/custom_checkers.html
# This is an example of a Pylint AST checker and should not be registered to use
# In an AST (abstract syntax tree) checker, the code will be represented as nodes of a tree
# We will use the astroid library: https://astroid.readthedocs.io/en/latest/api/general.html to visit and leave nodes
# Libraries needed for an AST checker
from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from pylint.lint import PyLinter


# Basic example of a Pylint AST (astract syntax tree) checker
# Checks for variables that have been reassigned in a function. If it finds a reassigned variable, it will throw an error
class NoReassignmentChecker(BaseChecker):
    __implements__ = IAstroidChecker

    # Checker name
    name = "no-reassigned-variable"
    # Set priority to -1
    priority = -1
    # Message dictionary
    msgs = {
        # message-id, consists of a letter and numbers
        # Letter will be one of following letters (C=Convention, W=Warning, E=Error, F=Fatal, R=Refactoring)
        # Numbers need to be unique and in-between 9000-9999
        # Check https://baseplate.readthedocs.io/en/stable/linters/index.html#custom-checkers-list
        # for numbers that are already in use
        "W9001": (
            # displayed-message shown to user
            "Reassigned variable found.",
            # message-symbol used as alias for message-id
            "reassigned-variable",
            # message-help shown to user when calling pylint --help-msg
            "Ensure variables are not reassigned.",
        )
    }

    def __init__(self, linter: PyLinter = None):
        super().__init__(linter)
        self.variables: set = set()

    # The following two methods are called for us by pylint/astroid
    # The linter walks through the tree, visiting and leaving desired nodes
    # Methods should start with visit_ or leave_ followed by lowercase class name of nodes
    # List of available nodes: https://astroid.readthedocs.io/en/latest/api/astroid.nodes.html

    # Visit the Assign node: https://astroid.readthedocs.io/en/latest/api/astroid.nodes.html#astroid.nodes.Assign
    def visit_assign(self, node: nodes) -> None:
        for variable in node.targets:
            if variable.name not in self.variables:
                self.variables.add(variable.name)
            else:
                self.add_message("non-unique-variable", node=node)

    # Leave the FunctionDef node: https://astroid.readthedocs.io/en/latest/api/astroid.nodes.html#astroid.nodes.FunctionDef
    def leave_functiondef(self, node: nodes) -> nodes:
        self.variables = set()
        return node



import base64
import logging
import sys
import time

from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import Mapping
from typing import Optional

import pyramid.events
import pyramid.request
import pyramid.tweens
import webob.request

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram
from pyramid.config import Configurator
from pyramid.registry import Registry
from pyramid.request import Request
from pyramid.response import Response

from baseplate import Baseplate
from baseplate import RequestContext
from baseplate import Span
from baseplate import TraceInfo
from baseplate.lib.edgecontext import EdgeContextFactory
from baseplate.lib.prometheus_metrics import default_latency_buckets
from baseplate.lib.prometheus_metrics import default_size_buckets
from baseplate.lib.prometheus_metrics import getHTTPSuccessLabel
from baseplate.thrift.ttypes import IsHealthyProbe


logger = logging.getLogger(__name__)


class SpanFinishingAppIterWrapper:
    """Wrapper for Response.app_iter that finishes the span when the iterator is done.

    The WSGI spec expects applications to return an iterable object. In the
    common case, the iterable is a single-item list containing a byte string of
    the full response. However, if the application wants to stream a response
    back to the client (e.g. it's sending a lot of data, or it wants to get
    some bytes on the wire really quickly before some database calls finish)
    the iterable can take a while to finish iterating.

    This wrapper allows us to keep the server span open until the iterable is
    finished even though our view callable returned long ago.

    """

    def __init__(self, span: Span, app_iter: Iterable[bytes]) -> None:
        self.span = span
        self.app_iter = iter(app_iter)

    def __iter__(self) -> Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        try:
            return next(self.app_iter)
        except StopIteration:
            self.span.finish()
            raise
        except:  # noqa: E722
            self.span.finish(exc_info=sys.exc_info())
            raise

    def close(self) -> None:
        if hasattr(self.app_iter, "close"):
            self.app_iter.close()  # type: ignore


PROM_NAMESPACE = "http_server"

HISTOGRAM_LABELS = [
    "http_method",
    "http_endpoint",
    "http_success",
]
REQUEST_LATENCY = Histogram(
    f"{PROM_NAMESPACE}_latency_seconds",
    "Time spent processing requests",
    HISTOGRAM_LABELS,
    buckets=default_latency_buckets,
)
REQUEST_SIZE = Histogram(
    f"{PROM_NAMESPACE}_request_size_bytes",
    "Size of incoming requests in bytes",
    HISTOGRAM_LABELS,
    buckets=default_size_buckets,
)
RESPONSE_SIZE = Histogram(
    f"{PROM_NAMESPACE}_response_size_bytes",
    "Size of outgoing responses in bytes",
    HISTOGRAM_LABELS,
    buckets=default_size_buckets,
)
REQUESTS_TOTAL = Counter(
    f"{PROM_NAMESPACE}_requests_total",
    "Total number of request handled",
    [
        *HISTOGRAM_LABELS,
        "http_response_code",
    ],
)
ACTIVE_REQUESTS = Gauge(
    f"{PROM_NAMESPACE}_active_requests",
    "Current requests in flight",
    [
        "http_method",
        "http_endpoint",
    ],
)


def _make_baseplate_tween(
    handler: Callable[[Request], Response], _registry: Registry
) -> Callable[[Request], Response]:
    def baseplate_tween(request: Request) -> Response:
        response: Optional[Response] = None

        try:
            response = handler(request)
            if request.span:
                request.span.set_tag("http.response_length", response.content_length)
        except:  # noqa: E722
            if hasattr(request, "span") and request.span:
                request.span.finish(exc_info=sys.exc_info())
            raise
        else:
            if request.span:
                request.span.set_tag("http.status_code", response.status_code)
                content_length = response.content_length
                response.app_iter = SpanFinishingAppIterWrapper(request.span, response.app_iter)
                response.content_length = content_length
        finally:
            manually_close_request_metrics(request, response)

            # avoid a reference cycle
            request.start_server_span = None
        return response

    return baseplate_tween


def manually_close_request_metrics(request: Request, response: Optional[Response] = None) -> None:
    """
    Close the request metrics and track the remaining bits of the request

    This is called both from the tween, but also available as a mechanism for pyramid scripting
    to mark that the request has finished.
    """
    # ensure any active counters have been incremented before decrementing them and tracking the
    # rest of the request
    if getattr(request, "reddit_prom_metrics_enabled", False):
        http_endpoint = ""
        if (
            hasattr(request, "reddit_tracked_endpoint")
            and request.reddit_tracked_endpoint is not None
        ):
            http_endpoint = request.reddit_tracked_endpoint
        elif request.matched_route:
            http_endpoint = (
                request.matched_route.pattern
                if (hasattr(request.matched_route, "pattern") and request.matched_route.pattern)
                else request.matched_route.name
            )
        else:
            http_endpoint = "404"

        http_method = request.method.lower()
        http_response_code = ""

        if sys.exc_info() == (None, None, None):
            if response:
                http_success = (
                    getHTTPSuccessLabel(int(response.status_code)) if response else "false"
                )
                http_response_code = response.status_code if response else ""
            else:
                http_success = "true"
                http_response_code = "200"
        else:
            http_success = "false"

        histogram_labels = {
            "http_method": http_method,
            "http_endpoint": http_endpoint,
            "http_success": http_success,
        }

        ACTIVE_REQUESTS.labels(http_method=http_method, http_endpoint=http_endpoint).dec()
        REQUESTS_TOTAL.labels(
            **{
                **histogram_labels,
                "http_response_code": http_response_code,
            }
        ).inc()

        if hasattr(request, "reddit_start_time") and request.reddit_start_time is not None:
            # note this is set in _on_new_request
            REQUEST_LATENCY.labels(**histogram_labels).observe(
                time.perf_counter() - request.reddit_start_time
            )

        # do it this way for tests and for services that bastardize the request object
        # for script execution where this may not be set
        if hasattr(request, "content_length") and request.content_length is not None:
            REQUEST_SIZE.labels(**histogram_labels).observe(request.content_length)

        # response may not be set if this handler is called from a pyramid script handler
        if response:
            if hasattr(response, "content_length") and response.content_length is not None:
                RESPONSE_SIZE.labels(**histogram_labels).observe(response.content_length)

        # avoid missing a secondary request if the same request object is re-used in scripting
        request.reddit_prom_metrics_enabled = False
        request.reddit_start_time = None
        request.reddit_tracked_endpoint = None
    else:
        logger.debug(
            "Request metrics attempted to be closed but were never opened, no metrics will be tracked"
        )


class BaseplateEvent:
    def __init__(self, request: Request):
        self.request = request


class ServerSpanInitialized(BaseplateEvent):
    """Event that Baseplate fires after creating the ServerSpan for a Request.

    This event will be emitted before the Request is passed along to it's
    handler.  Baseplate initializes the ServerSpan in response to a
    :py:class:`pyramid.events.ContextFound` event emitted by Pyramid so while
    we can guarantee what Baseplate has done when this event is emitted, we
    cannot guarantee that any other subscribers to
    :py:class:`pyramid.events.ContextFound` have been called or not.
    """


class HeaderTrustHandler:
    """Abstract class used by :py:class:`BaseplateConfigurator` to validate headers.

    See :py:class:`StaticTrustHandler` for the default implementation.
    """

    def should_trust_trace_headers(self, request: Request) -> bool:
        """Return whether baseplate should parse the trace headers from the inbound request.

        :param request: The request

        :returns: Whether baseplate should parse the trace headers from the inbound request.
        """
        raise NotImplementedError

    def should_trust_edge_context_payload(self, request: Request) -> bool:
        """Return whether baseplate should trust the edge context headers from the inbound request.

        :param request: The request

        :returns: Whether baseplate should trust the inbound edge context headers
        """
        raise NotImplementedError


class StaticTrustHandler(HeaderTrustHandler):
    """Default implementation for handling headers.

    This class is created automatically by BaseplateConfigurator unless you
    supply your own HeaderTrustHandler

    :param trust_headers:
        Whether or not to trust trace and edge context headers from
        inbound requests. This value will be returned by should_trust_trace_headers and
        should_trust_edge_context_payload.

    .. warning::

        Do not set ``trust_headers`` to ``True`` unless you are sure your
        application is only accessible by trusted sources (usually backend-only
        services).
    """

    def __init__(self, trust_headers: bool = False):
        self.trust_headers = trust_headers

    def should_trust_trace_headers(self, request: Request) -> bool:
        return self.trust_headers

    def should_trust_edge_context_payload(self, request: Request) -> bool:
        return self.trust_headers


# pylint: disable=too-many-ancestors
class BaseplateRequest(RequestContext, pyramid.request.Request):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        context_config = kwargs.pop("context_config", None)
        RequestContext.__init__(self, context_config=context_config)
        pyramid.request.Request.__init__(self, *args, **kwargs)


class RequestFactory:
    def __init__(self, baseplate: Baseplate):
        self.baseplate = baseplate

    def __call__(self, environ: Dict[str, str]) -> BaseplateRequest:
        return BaseplateRequest(environ, context_config=self.baseplate._context_config)

    def blank(self, path: str) -> BaseplateRequest:
        environ = webob.request.environ_from_url(path)
        return BaseplateRequest(environ, context_config=self.baseplate._context_config)


class BaseplateConfigurator:
    """Configuration extension to integrate Baseplate into Pyramid.

    :param baseplate: The Baseplate instance for your application.
    :param edge_context_factory: A configured factory for handling edge request
        context.
    :param header_trust_handler: An object which will be used to verify whether
        baseplate should parse the request context headers, for example trace ids.
        See StaticTrustHandler for the default implementation.
    """

    def __init__(
        self,
        baseplate: Baseplate,
        edge_context_factory: Optional[EdgeContextFactory] = None,
        header_trust_handler: Optional[HeaderTrustHandler] = None,
    ):
        self.baseplate = baseplate
        self.edge_context_factory = edge_context_factory
        self.header_trust_handler = header_trust_handler or StaticTrustHandler(trust_headers=False)

    def _on_application_created(self, event: pyramid.events.ApplicationCreated) -> None:
        # attach the baseplate object to the application the server gets
        event.app.baseplate = self.baseplate

    def _on_new_request(self, event: pyramid.events.ContextFound) -> None:
        request = event.request
        endpoint = ""

        if request.matched_route:
            endpoint = (
                request.matched_route.pattern
                if (hasattr(request.matched_route, "pattern") and request.matched_route.pattern)
                else request.matched_route.name
            )
        else:
            endpoint = "404"
        request.reddit_prom_metrics_enabled = True
        request.reddit_tracked_endpoint = endpoint
        request.reddit_start_time = time.perf_counter()
        ACTIVE_REQUESTS.labels(http_method=request.method.lower(), http_endpoint=endpoint).inc()

        # this request didn't match a route we know
        if not request.matched_route:
            return

        trace_info = None
        if self.header_trust_handler.should_trust_trace_headers(request):
            try:
                trace_info = self._get_trace_info(request.headers)
            except (KeyError, ValueError):
                pass

        if self.header_trust_handler.should_trust_edge_context_payload(request):
            edge_payload: Optional[bytes]
            try:
                edge_payload_str = request.headers["X-Edge-Request"]
                edge_payload = base64.b64decode(edge_payload_str.encode())
            except (KeyError, ValueError):
                edge_payload = None

            request.raw_edge_context = edge_payload
            if self.edge_context_factory:
                request.edge_context = self.edge_context_factory.from_upstream(edge_payload)

        span = self.baseplate.make_server_span(
            request,
            name=request.matched_route.name,
            trace_info=trace_info,
        )
        span.set_tag("protocol", "http")
        span.set_tag("http.url", request.url)
        span.set_tag("http.method", request.method)
        span.set_tag("peer.ipv4", request.remote_addr)
        span.start()

        request.registry.notify(ServerSpanInitialized(request))

    def _get_trace_info(self, headers: Mapping[str, str]) -> TraceInfo:
        sampled = bool(headers.get("X-Sampled") == "1")
        flags = headers.get("X-Flags", None)
        return TraceInfo.from_upstream(
            headers["X-Trace"],
            headers["X-Parent"],
            headers["X-Span"],
            sampled,
            int(flags) if flags is not None else None,
        )

    def includeme(self, config: Configurator) -> None:
        config.set_request_factory(RequestFactory(self.baseplate))
        config.add_subscriber(self._on_new_request, pyramid.events.ContextFound)
        config.add_subscriber(self._on_application_created, pyramid.events.ApplicationCreated)

        # Position of the tween is important. We need it to cover all code
        # that can written in the app. This means that it should be above
        # (wrap) both the exception view handling tween (EXCVIEW) and the
        # request handling code (MAIN).
        #
        # The final order ends up being:
        # 1. Request ingress (tweens.INGRESS)
        # 2. baseplate tween
        # 3. Exception view handler (tweens.EXCVIEW)
        # 4. App handler code (tweens.MAIN)
        config.add_tween(
            "baseplate.frameworks.pyramid._make_baseplate_tween", over=pyramid.tweens.EXCVIEW
        )


def get_is_healthy_probe(request: Request) -> int:
    """Get the thrift enum value of the probe used in http is_healthy request."""
    code = request.params.get("type", str(IsHealthyProbe.READINESS))
    try:
        return int(code)
    except ValueError:
        pass
    # If it's not an int, try to find it in the enum map instead.
    try:
        return IsHealthyProbe._NAMES_TO_VALUES[code.upper()]
    except KeyError:
        logger.warning(
            "Unrecognized health check type %s, fallback to READINESS",
            code,
        )
        return IsHealthyProbe.READINESS



try:
    # nullcontext is only available in Python 3.7+
    from contextlib import nullcontext as does_not_raise
except ImportError:
    from contextlib import contextmanager

    @contextmanager
    def does_not_raise():
        yield


__all__ = [
    "does_not_raise",
]



# sphinx docs config
#
# This file is execfile()d with the current directory set to its
# containing dir.
import os
import sys

import alabaster

# where to find the modules to import for autodoc
sys.path.insert(0, os.path.abspath(".."))

# -- General configuration ------------------------------------------------
# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "alabaster",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3.7", None),
    "pyramid": (
        "https://docs.pylonsproject.org/projects/pyramid/en/1.9-branch",
        None,
    ),
    "cassandra": ("https://datastax.github.io/python-driver/", None),
    "pymemcache": ("https://pymemcache.readthedocs.io/en/latest/", None),
    "kazoo": ("https://kazoo.readthedocs.io/en/latest/", None),
    "kombu": ("https://kombu.readthedocs.io/en/latest/", None),
    "redis": ("https://redis-py.readthedocs.io/en/latest/", None),
    "sqlalchemy": ("https://docs.sqlalchemy.org/en/13/", None),
    "requests": ("https://requests.readthedocs.io/en/stable/", None),
    "sentry_sdk": ("https://getsentry.github.io/sentry-python/", None),
}

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
source_encoding = "utf-8"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Baseplate.py"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["docs/html"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "friendly"

# -- Options for HTML output ----------------------------------------------

html_theme_path = [alabaster.get_path()]
html_static_path = ["images"]

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "alabaster"

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title =

# which templates to put in the sidebar.  we're just removing the relations
# section from the defaults here, that's "next article" and "previous article"
html_sidebars = {
    "**": ["about.html", "searchbox.html", "navigation.html"]
}

html_theme_options = {
    "description": "Reddit's Python Service Framework",
    "github_button": False,
    "github_repo": "baseplate.py",
    "github_user": "reddit",
    "github_banner": True,
    "logo": "baseplate.png",
    "logo_name": True,
    "show_powered_by": False,
    "show_related": False,
    "show_relbars": True,
    "page_width": "960px",
}

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "images/favicon.png"

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# -- Autodoc --
autodoc_member_order = "bysource"



#
# Autogenerated by Thrift Compiler (0.14.1)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:slots
#
import logging
import sys

from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.Thrift import TFrozenDict
from thrift.Thrift import TMessageType
from thrift.Thrift import TProcessor
from thrift.Thrift import TType
from thrift.transport import TTransport
from thrift.TRecursive import fix_spec

from .ttypes import *

all_structs = []


class Iface(object):
    """
    The base for any baseplate-based service.

    Your service should inherit from this one so that common tools can interact
    with any expected interfaces.

    DEPRECATED: Please migrate to BaseplateServiceV2.


    """

    def is_healthy(self):
        """
        Return whether or not the service is healthy.

        The healthchecker (baseplate.server.healthcheck) expects this endpoint to
        exist so it can determine your service's health.

        This should return True if the service is healthy. If the service is
        unhealthy, it can return False or raise an exception.


        """
        pass


class Client(Iface):
    """
    The base for any baseplate-based service.

    Your service should inherit from this one so that common tools can interact
    with any expected interfaces.

    DEPRECATED: Please migrate to BaseplateServiceV2.


    """

    def __init__(self, iprot, oprot=None):
        self._iprot = self._oprot = iprot
        if oprot is not None:
            self._oprot = oprot
        self._seqid = 0

    def is_healthy(self):
        """
        Return whether or not the service is healthy.

        The healthchecker (baseplate.server.healthcheck) expects this endpoint to
        exist so it can determine your service's health.

        This should return True if the service is healthy. If the service is
        unhealthy, it can return False or raise an exception.


        """
        self.send_is_healthy()
        return self.recv_is_healthy()

    def send_is_healthy(self):
        self._oprot.writeMessageBegin("is_healthy", TMessageType.CALL, self._seqid)
        args = is_healthy_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_is_healthy(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = is_healthy_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(
            TApplicationException.MISSING_RESULT, "is_healthy failed: unknown result"
        )


class Processor(Iface, TProcessor):
    def __init__(self, handler):
        self._handler = handler
        self._processMap = {}
        self._processMap["is_healthy"] = Processor.process_is_healthy
        self._on_message_begin = None

    def on_message_begin(self, func):
        self._on_message_begin = func

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if self._on_message_begin:
            self._on_message_begin(name, type, seqid)
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(
                TApplicationException.UNKNOWN_METHOD, "Unknown function %s" % (name)
            )
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_is_healthy(self, seqid, iprot, oprot):
        args = is_healthy_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = is_healthy_result()
        try:
            result.success = self._handler.is_healthy()
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception("TApplication exception in handler")
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception("Unexpected exception in handler")
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, "Internal error")
        oprot.writeMessageBegin("is_healthy", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES


class is_healthy_args(object):

    __slots__ = ()

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin("is_healthy_args")
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, getattr(self, key)) for key in self.__slots__]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            my_val = getattr(self, attr)
            other_val = getattr(other, attr)
            if my_val != other_val:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)


all_structs.append(is_healthy_args)
is_healthy_args.thrift_spec = ()


class is_healthy_result(object):
    """
    Attributes:
     - success

    """

    __slots__ = ("success",)

    def __init__(
        self,
        success=None,
    ):
        self.success = success

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.BOOL:
                    self.success = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin("is_healthy_result")
        if self.success is not None:
            oprot.writeFieldBegin("success", TType.BOOL, 0)
            oprot.writeBool(self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, getattr(self, key)) for key in self.__slots__]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            my_val = getattr(self, attr)
            other_val = getattr(other, attr)
            if my_val != other_val:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)


all_structs.append(is_healthy_result)
is_healthy_result.thrift_spec = (
    (
        0,
        TType.BOOL,
        "success",
        None,
        None,
    ),  # 0
)
fix_spec(all_structs)
del all_structs



from baseplate.lib.ratelimit.ratelimit import RateLimiter
from baseplate.lib.ratelimit.ratelimit import RateLimiterContextFactory
from baseplate.lib.ratelimit.ratelimit import RateLimitExceededException


__all__ = ["RateLimiter", "RateLimitExceededException", "RateLimiterContextFactory"]



import unittest

from unittest import mock

from baseplate import LocalSpan
from baseplate import ServerSpan
from baseplate import Span
from baseplate.lib.metrics import Batch
from baseplate.lib.metrics import Client
from baseplate.lib.metrics import Counter
from baseplate.lib.metrics import Timer
from baseplate.observers.metrics import MetricsBaseplateObserver
from baseplate.observers.metrics import MetricsClientSpanObserver
from baseplate.observers.metrics import MetricsLocalSpanObserver
from baseplate.observers.metrics import MetricsServerSpanObserver


class TestException(Exception):
    pass


class ObserverTests(unittest.TestCase):
    def test_add_to_context(self):
        mock_client = mock.Mock(spec=Client)
        mock_batch = mock_client.batch.return_value
        mock_context = mock.Mock()
        mock_server_span = mock.Mock(spec=ServerSpan)
        mock_server_span.name = "name"

        observer = MetricsBaseplateObserver(mock_client)
        observer.on_server_span_created(mock_context, mock_server_span)

        self.assertEqual(mock_context.metrics, mock_batch)
        self.assertEqual(mock_server_span.register.call_count, 1)


class ServerSpanObserverTests(unittest.TestCase):
    def test_server_span_events(self):
        mock_batch = mock.Mock(spec=Batch)
        mock_timer = mock.Mock(spec=Timer)
        mock_counter = mock.Mock(spec=Counter)
        mock_batch.timer.return_value = mock_timer
        mock_batch.counter.return_value = mock_counter

        mock_server_span = mock.Mock(spec=ServerSpan)
        mock_server_span.name = "request_name"

        observer = MetricsServerSpanObserver(mock_batch, mock_server_span)

        observer.on_start()
        self.assertEqual(mock_timer.start.call_count, 1)

        observer.on_incr_tag("test", delta=1)
        self.assertEqual(mock_counter.increment.call_count, 1)

        mock_child_span = mock.Mock()
        mock_child_span.name = "example"
        observer.on_child_span_created(mock_child_span)
        self.assertEqual(mock_child_span.register.call_count, 1)

        observer.on_finish(exc_info=None)
        self.assertEqual(mock_timer.stop.call_count, 1)
        self.assertEqual(mock_batch.flush.call_count, 1)


class ClientSpanObserverTests(unittest.TestCase):
    def test_metrics(self):
        mock_timer = mock.Mock(spec=Timer)
        mock_counter = mock.Mock(spec=Counter)
        mock_batch = mock.Mock(spec=Batch)
        mock_batch.timer.return_value = mock_timer
        mock_batch.counter.return_value = mock_counter

        mock_client_span = mock.Mock(spec=Span)
        mock_client_span.name = "example"

        observer = MetricsClientSpanObserver(mock_batch, mock_client_span)
        self.assertEqual(mock_batch.timer.call_count, 1)
        self.assertEqual(mock_batch.timer.call_args, mock.call("clients.example"))

        observer.on_start()
        self.assertEqual(mock_timer.start.call_count, 1)

        observer.on_incr_tag("test", delta=1)
        mock_counter.increment.assert_called()
        mock_counter.reset_mock()

        observer.on_finish(exc_info=None)
        self.assertEqual(mock_timer.stop.call_count, 1)
        self.assertEqual(mock_counter.increment.call_count, 1)

        mock_counter.reset_mock()
        observer.on_log(name="error.object", payload=TestException())
        self.assertEqual(mock_counter.increment.call_count, 1)
        self.assertEqual(mock_batch.counter.call_args, mock.call("errors.TestException"))


class LocalSpanObserverTests(unittest.TestCase):
    def test_metrics(self):
        mock_timer = mock.Mock(spec=Timer)
        mock_counter = mock.Mock(spec=Counter)
        mock_batch = mock.Mock(spec=Batch)
        mock_batch.timer.return_value = mock_timer
        mock_batch.counter.return_value = mock_counter

        mock_local_span = mock.Mock(spec=LocalSpan)
        mock_local_span.name = "example"
        mock_local_span.component_name = "some_component"

        observer = MetricsLocalSpanObserver(mock_batch, mock_local_span)
        self.assertEqual(mock_batch.timer.call_count, 1)
        self.assertEqual(mock_batch.timer.call_args, mock.call("some_component.example"))

        observer.on_start()
        self.assertEqual(mock_timer.start.call_count, 1)

        observer.on_incr_tag("test", delta=1)
        mock_counter.increment.assert_called()
        mock_counter.reset_mock()

        observer.on_finish(exc_info=None)
        self.assertEqual(mock_timer.stop.call_count, 1)

    def test_spans_under_local_spans(self):
        mock_batch = mock.Mock(spec=Batch)

        mock_local_span = mock.Mock(spec=LocalSpan)
        mock_local_span.name = "example"
        mock_local_span.component_name = "some_component"

        mock_nested_span = mock.Mock(spec=LocalSpan)
        mock_nested_span.name = "nested"
        mock_nested_span.component_name = "some_component2"

        observer = MetricsLocalSpanObserver(mock_batch, mock_local_span)
        observer.on_child_span_created(mock_nested_span)

        self.assertEqual(mock_nested_span.register.call_count, 1)



from redis import ConnectionPool

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.clients.redis import MonitoredRedisConnection
from baseplate.clients.redis import RedisContextFactory
from baseplate.lib.ratelimit.backends import _get_current_bucket
from baseplate.lib.ratelimit.backends import RateLimitBackend


class RedisRateLimitBackendContextFactory(ContextFactory):
    """RedisRateLimitBackend context factory.

    :param redis_pool: The redis pool to back this ratelimit.
    :param prefix: A prefix to add to keys during rate limiting.
        This is useful if you will have two different rate limiters that will
        receive the same keys.

    """

    def __init__(self, redis_pool: ConnectionPool, prefix: str = "rl:"):
        self.redis_context_factory = RedisContextFactory(redis_pool)
        self.prefix = prefix

    def make_object_for_context(self, name: str, span: Span) -> "RedisRateLimitBackend":
        redis = self.redis_context_factory.make_object_for_context(name, span)
        return RedisRateLimitBackend(redis, prefix=self.prefix)


class RedisRateLimitBackend(RateLimitBackend):
    """A Redis backend for rate limiting.

    :param redis: An instance of
        :py:class:`baseplate.clients.redis.MonitoredRedisConnection`.
    :param prefix: A prefix to add to keys during rate limiting.
        This is useful if you will have two different rate limiters that will
        receive the same keys.

    """

    def __init__(self, redis: MonitoredRedisConnection, prefix: str = "rl:"):
        self.redis = redis
        self.prefix = prefix

    def consume(self, key: str, amount: int, allowance: int, interval: int) -> bool:
        """Consume the given `amount` from the allowance for the given `key`.

        This will return true if the `key` remains below the `allowance`
        after consuming the given `amount`.

        :param key: The name of the rate limit bucket to consume from.
        :param amount: The amount to consume from the rate limit bucket.
        :param allowance: The maximum allowance for the rate limit bucket.
        :param interval: The interval to reset the allowance.

        """
        current_bucket = _get_current_bucket(interval)
        key = self.prefix + key + current_bucket
        ttl = interval * 2
        with self.redis.pipeline("ratelimit") as pipe:
            pipe.incr(key, amount)
            pipe.expire(key, time=ttl)
            responses = pipe.execute()
        count = responses[0]
        return count <= allowance



import gc
import logging
import os
import socket
import threading
import time

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import NoReturn
from typing import Optional

import gevent.events

from gevent.pool import Pool

from baseplate import _ExcInfo
from baseplate import Baseplate
from baseplate import BaseplateObserver
from baseplate import RequestContext
from baseplate import ServerSpan
from baseplate import ServerSpanObserver
from baseplate.lib import config
from baseplate.lib import metrics


REPORT_INTERVAL_SECONDS = 10
MAX_REQUEST_AGE = 60


logger = logging.getLogger(__name__)


class _Reporter:
    def report(self, batch: metrics.Batch) -> None:
        raise NotImplementedError


class _OpenConnectionsReporter(_Reporter):
    def __init__(self, pool: Pool):
        self.pool = pool

    def report(self, batch: metrics.Batch) -> None:
        batch.gauge("open_connections").replace(len(self.pool.greenlets))


class _ActiveRequestsObserver(BaseplateObserver, _Reporter):
    def __init__(self) -> None:
        self.live_requests: Dict[str, float] = {}

    def on_server_span_created(self, context: RequestContext, server_span: ServerSpan) -> None:
        observer = _ActiveRequestsServerSpanObserver(self, server_span.trace_id)
        server_span.register(observer)

    def report(self, batch: metrics.Batch) -> None:
        threshold = time.time() - MAX_REQUEST_AGE
        stale_requests = [
            trace_id
            for trace_id, start_time in self.live_requests.items()
            if start_time < threshold
        ]
        for stale_request_id in stale_requests:
            self.live_requests.pop(stale_request_id, None)

        batch.gauge("active_requests").replace(len(self.live_requests))


class _ActiveRequestsServerSpanObserver(ServerSpanObserver):
    def __init__(self, reporter: _ActiveRequestsObserver, trace_id: str):
        self.reporter = reporter
        self.trace_id = trace_id

    def on_start(self) -> None:
        self.reporter.live_requests[self.trace_id] = time.time()

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        self.reporter.live_requests.pop(self.trace_id, None)


class _BlockedGeventHubReporter(_Reporter):
    def __init__(self, max_blocking_time: int):
        gevent.events.subscribers.append(self._on_gevent_event)
        gevent.config.monitor_thread = True
        gevent.config.max_blocking_time = max_blocking_time
        gevent.get_hub().start_periodic_monitoring_thread()

        self.times_blocked: List[int] = []

    def _on_gevent_event(self, event: Any) -> None:
        if isinstance(event, gevent.events.EventLoopBlocked):
            self.times_blocked.append(event.blocking_time)

    def report(self, batch: metrics.Batch) -> None:
        # gevent events come in on another thread. we're relying on the GIL to
        # keep us from shenanigans here and we swap things out semi-safely to
        # ensure minimal lost data.
        times_blocked = self.times_blocked
        self.times_blocked = []

        for time_blocked in times_blocked:
            batch.timer("hub_blocked").send(time_blocked)


class _GCStatsReporter(_Reporter):
    def report(self, batch: metrics.Batch) -> None:
        for generation, stats in enumerate(gc.get_stats()):
            for name, value in stats.items():
                gauge = batch.gauge(f"gc.{name}", tags={"generation": generation})
                gauge.replace(value)


class _GCTimingReporter(_Reporter):
    def __init__(self) -> None:
        gc.callbacks.append(self._on_gc_event)

        self.gc_durations: List[float] = []
        self.current_gc_start: Optional[float] = None

    def _on_gc_event(self, phase: str, _info: Dict[str, Any]) -> None:
        if phase == "start":
            self.current_gc_start = time.time()
        elif phase == "stop":
            if self.current_gc_start:
                elapsed = time.time() - self.current_gc_start
                self.current_gc_start = None
                self.gc_durations.append(elapsed)

    def report(self, batch: metrics.Batch) -> None:
        gc_durations = self.gc_durations
        self.gc_durations = []

        for gc_duration in gc_durations:
            batch.timer("gc.elapsed").send(gc_duration)


class _BaseplateReporter(_Reporter):
    def __init__(self, reporters: Dict[str, Callable[[Any], None]]):
        self.reporters = reporters

    def report(self, batch: metrics.Batch) -> None:
        for name, reporter in self.reporters.items():
            try:
                batch.base_tags["client"] = name
                reporter(batch)
            except Exception as exc:
                logger.exception("Error generating client metrics: %s: %s", name, exc)
            finally:
                del batch.base_tags["client"]


class _RefCycleReporter(_Reporter):
    def __init__(self, root: str):
        assert os.path.isdir(root), f"{root} is not a directory"
        self.root = root

        # test that this is available up front
        import objgraph

        del objgraph

        logger.warning("Disabling automatic garbage collection to watch for reference cycles.")
        gc.disable()

    def report(self, batch: metrics.Batch) -> None:  # pylint: disable=unused-argument
        # run a garbage collection but keep everything we found in gc.garbage
        gc.set_debug(gc.DEBUG_SAVEALL)
        gc.collect()

        if gc.garbage:
            import objgraph

            logger.warning("%d objects garbage collected. Writing objgraph...", len(gc.garbage))
            objgraph.show_backrefs(
                gc.garbage, filename=f"{self.root}/backrefs-{int(time.time())}.png"
            )

            # clean out the garbage altogether
            gc.garbage.clear()
            gc.set_debug(0)
            gc.collect()
        else:
            logger.debug("No garbage yet!")


def _report_runtime_metrics_periodically(
    metrics_client: metrics.Client, reporters: List[_Reporter]
) -> NoReturn:
    hostname = socket.gethostname()
    pid = str(os.getpid())

    while True:
        now = time.time()
        time_since_last_report = now % REPORT_INTERVAL_SECONDS
        time_until_next_report = REPORT_INTERVAL_SECONDS - time_since_last_report
        time.sleep(time_until_next_report)

        try:
            with metrics_client.batch() as batch:
                batch.namespace += b".runtime"
                batch.base_tags["hostname"] = hostname
                batch.base_tags["PID"] = pid

                for reporter in reporters:
                    try:
                        reporter.report(batch)
                    except Exception as exc:
                        logger.debug(
                            "Error generating server metrics: %s: %s",
                            reporter.__class__.__name__,
                            exc,
                        )
        except Exception as exc:
            logger.debug("Error while sending server metrics: %s", exc)


def start(server_config: Dict[str, str], application: Any, pool: Pool) -> None:
    baseplate: Baseplate = getattr(application, "baseplate", None)
    if not baseplate or not baseplate._metrics_client:
        logger.info("No metrics client configured. Server metrics will not be sent.")
        return

    cfg = config.parse_config(
        server_config,
        {
            "monitoring": {
                "blocked_hub": config.Optional(config.Timespan, default=None),
                "concurrency": config.Optional(config.Boolean, default=True),
                "connection_pool": config.Optional(config.Boolean, default=False),
                "gc": {
                    "stats": config.Optional(config.Boolean, default=True),
                    "timing": config.Optional(config.Boolean, default=False),
                    "refcycle": config.Optional(config.String, default=None),
                },
            }
        },
    )

    reporters: List[_Reporter] = []

    if cfg.monitoring.concurrency:
        reporters.append(_OpenConnectionsReporter(pool))
        observer = _ActiveRequestsObserver()
        reporters.append(observer)
        baseplate.register(observer)

    if cfg.monitoring.connection_pool:
        reporters.append(_BaseplateReporter(baseplate.get_runtime_metric_reporters()))

    if cfg.monitoring.blocked_hub is not None:
        try:
            reporters.append(_BlockedGeventHubReporter(cfg.monitoring.blocked_hub.total_seconds()))
        except Exception as exc:
            logger.info("monitoring.blocked_hub disabled: %s", exc)

    if cfg.monitoring.gc.stats:
        try:
            reporters.append(_GCStatsReporter())
        except Exception as exc:
            logger.info("monitoring.gc.stats disabled: %s", exc)

    if cfg.monitoring.gc.timing:
        try:
            reporters.append(_GCTimingReporter())
        except Exception as exc:
            logger.info("monitoring.gc.timing disabled: %s", exc)

    if cfg.monitoring.gc.refcycle:
        try:
            reporters.append(_RefCycleReporter(cfg.monitoring.gc.refcycle))
        except Exception as exc:
            logger.info("monitoring.gc.refcycle disabled: %s", exc)

    thread = threading.Thread(
        name="Server Monitoring",
        target=_report_runtime_metrics_periodically,
        args=(application.baseplate._metrics_client, reporters),
    )
    thread.daemon = True
    thread.start()



import logging
import queue
import socket
import time

from typing import Any
from typing import Callable
from typing import Dict
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import TYPE_CHECKING

import kombu

from gevent.server import StreamServer
from kombu.mixins import ConsumerMixin
from kombu.transport.virtual import Channel
from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram

from baseplate import Baseplate
from baseplate import RequestContext
from baseplate.clients.kombu import KombuSerializer
from baseplate.lib.prometheus_metrics import default_latency_buckets
from baseplate.server.queue_consumer import HealthcheckCallback
from baseplate.server.queue_consumer import make_simple_healthchecker
from baseplate.server.queue_consumer import MessageHandler
from baseplate.server.queue_consumer import PumpWorker
from baseplate.server.queue_consumer import QueueConsumerFactory


class AmqpConsumerPrometheusLabels(NamedTuple):
    amqp_address: str
    amqp_virtual_host: str
    amqp_exchange_name: str
    amqp_routing_key: str


AMQP_PROCESSING_TIME = Histogram(
    "amqp_consumer_message_processing_time_seconds",
    "latency histogram of how long it takes to process a message",
    AmqpConsumerPrometheusLabels._fields + ("amqp_success",),
    buckets=default_latency_buckets,
)

AMQP_PROCESSED_TOTAL = Counter(
    "amqp_consumer_messages_processed_total",
    "total count of messages processed by this host",
    AmqpConsumerPrometheusLabels._fields + ("amqp_success",),
)

AMQP_ACTIVE_MESSAGES = Gauge(
    "amqp_consumer_active_messages",
    "gauge that reflects the number of messages currently being processed",
    AmqpConsumerPrometheusLabels._fields,
)


if TYPE_CHECKING:
    WorkQueue = queue.Queue[kombu.Message]  # pylint: disable=unsubscriptable-object
else:
    WorkQueue = queue.Queue

logger = logging.getLogger(__name__)


Handler = Callable[[RequestContext, Any, kombu.Message], None]
ErrorHandler = Callable[[RequestContext, Any, kombu.Message, Exception], None]


class FatalMessageHandlerError(Exception):
    """An error that signals that the queue process should exit.

    Raising an Exception that is a subclass of FatalMessageHandlerError will
    cause the KombuMessageHandler to re-raise the exception rather than swallowing
    it which will cause the handler thread/process to stop.  This, in turn, will
    gracefully shut down the QueueConsumerServer currently running.

    Exceptions of this nature should be reserved for errors that are due to
    problems with the environment rather than the message itself.  For example,
    a node that cannot get its AWS credentials.
    """


class KombuConsumerWorker(ConsumerMixin, PumpWorker):
    """Consumes messages from the given queues and pumps them into the internal work_queue.

    This class does not directly implement the abstract `run` command from
    PumpWorker because the ConsumerMixin class already defines it.
    """

    def __init__(
        self,
        connection: kombu.Connection,
        queues: Sequence[kombu.Queue],
        work_queue: WorkQueue,
        serializer: Optional[KombuSerializer] = None,
        **kwargs: Any,
    ):
        self.connection = connection
        self.queues = queues
        self.work_queue = work_queue
        self.serializer = serializer
        self.kwargs = kwargs

    def get_consumers(self, Consumer: kombu.Consumer, channel: Channel) -> Sequence[kombu.Consumer]:
        args = dict(queues=self.queues, on_message=self.work_queue.put, **self.kwargs)
        if self.serializer:
            args["accept"] = [self.serializer.name]
        return [Consumer(**args)]

    def stop(self) -> None:
        logger.debug("Closing KombuConsumerWorker.")
        # `should_stop` is an attribute of `ConsumerMixin`
        self.should_stop = True


class KombuMessageHandler(MessageHandler):
    def __init__(
        self,
        baseplate: Baseplate,
        name: str,
        handler_fn: Handler,
        error_handler_fn: Optional[ErrorHandler] = None,
    ):
        self.baseplate = baseplate
        self.name = name
        self.handler_fn = handler_fn
        self.error_handler_fn = error_handler_fn

    def handle(self, message: kombu.Message) -> None:
        start_time = time.perf_counter()
        prometheus_success = "true"
        prometheus_labels = AmqpConsumerPrometheusLabels(
            amqp_address=message.channel.connection.client.host,  # note: localhost will be translated to 127.0.0.1 by the library
            amqp_virtual_host=message.channel.connection.client.virtual_host,
            amqp_exchange_name=message.delivery_info.get("exchange", ""),
            amqp_routing_key=message.delivery_info.get("routing_key", ""),
        )

        context = self.baseplate.make_context_object()
        try:
            # We place the call to ``baseplate.make_server_span`` inside the
            # try/except block because we still want Baseplate to see and
            # handle the error (publish it to error reporting)
            with self.baseplate.make_server_span(
                context, self.name
            ) as span, AMQP_ACTIVE_MESSAGES.labels(
                **prometheus_labels._asdict()
            ).track_inprogress():
                delivery_info = message.delivery_info
                message_body = None
                message_body = message.decode()
                span.set_tag("kind", "consumer")
                span.set_tag("amqp.routing_key", delivery_info.get("routing_key", ""))
                span.set_tag("amqp.consumer_tag", delivery_info.get("consumer_tag", ""))
                span.set_tag("amqp.delivery_tag", delivery_info.get("delivery_tag", ""))
                span.set_tag("amqp.exchange", delivery_info.get("exchange", ""))
                self.handler_fn(context, message_body, message)
        except Exception as exc:
            prometheus_success = "false"
            logger.exception(
                "Unhandled error while trying to process a message.  The message "
                "has been returned to the queue broker."
            )
            if self.error_handler_fn:
                self.error_handler_fn(context, message_body, message, exc)
            else:
                message.requeue()

            if isinstance(exc, FatalMessageHandlerError):
                logger.info("Recieved a fatal error, terminating the server.")
                raise
        else:
            message.ack()
        finally:
            AMQP_PROCESSING_TIME.labels(
                **prometheus_labels._asdict(), amqp_success=prometheus_success
            ).observe(time.perf_counter() - start_time)
            AMQP_PROCESSED_TOTAL.labels(
                **prometheus_labels._asdict(), amqp_success=prometheus_success
            ).inc()


class KombuQueueConsumerFactory(QueueConsumerFactory):
    """Factory for running a :py:class:`~baseplate.server.queue_consumer.QueueConsumerServer` using Kombu.

    For simple cases where you just need a basic queue with all the default
    parameters for your message broker, you can use `KombuQueueConsumerFactory.new`.

    If you need more control, you can create the :py:class:`~kombu.Queue` s yourself and
    use the constructor directly.
    """

    def __init__(
        self,
        baseplate: Baseplate,
        name: str,
        connection: kombu.Connection,
        queues: Sequence[kombu.Queue],
        handler_fn: Handler,
        error_handler_fn: Optional[ErrorHandler] = None,
        health_check_fn: Optional[HealthcheckCallback] = None,
        serializer: Optional[KombuSerializer] = None,
        worker_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """`KombuQueueConsumerFactory` constructor.

        :param baseplate: The Baseplate set up for your consumer.
        :param exchange: The `kombu.Exchange` that you will bind your :py:class:`~kombu.Queue` s
            to.
        :param queues: List of  :py:class:`~kombu.Queue` s to consume from.
        :param queue_name: Name for your queue.
        :param routing_keys: List of routing keys that you will create :py:class:`~kombu.Queue` s
            to consume from.
        :param handler_fn: A function that will process an individual message from a queue.
        :param error_handler_fn: A function that will be called when an error is thrown
            while executing the `handler_fn`. This function will be responsible for calling
            `message.ack` or `message.requeue` as it will not be automatically called by
            `KombuMessageHandler`'s `handle` function.
        :param health_check_fn: A `baseplate.server.queue_consumer.HealthcheckCallback`
            function that can be used to customize your health check.
        :param serializer: A `baseplate.clients.kombu.KombuSerializer` that should
            be used to decode the messages you are consuming.
        :param worker_kwargs: A dictionary of keyword arguments used to create queue consumers.
        """
        self.baseplate = baseplate
        self.connection = connection
        self.queues = queues
        self.name = name
        self.handler_fn = handler_fn
        self.error_handler_fn = error_handler_fn
        self.health_check_fn = health_check_fn
        self.serializer = serializer
        self.worker_kwargs = worker_kwargs

    @classmethod
    def new(
        cls,
        baseplate: Baseplate,
        exchange: kombu.Exchange,
        connection: kombu.Connection,
        queue_name: str,
        routing_keys: Sequence[str],
        handler_fn: Handler,
        error_handler_fn: Optional[ErrorHandler] = None,
        health_check_fn: Optional[HealthcheckCallback] = None,
        serializer: Optional[KombuSerializer] = None,
        worker_kwargs: Optional[Dict[str, Any]] = None,
    ) -> "KombuQueueConsumerFactory":
        """Return a new `KombuQueueConsumerFactory`.

        This method will create the :py:class:`~kombu.Queue` s for you and is
        appropriate to use in simple cases where you just need a basic queue with
        all the default parameters for your message broker.

        :param baseplate: The Baseplate set up for your consumer.
        :param exchange: The `kombu.Exchange` that you will bind your
            :py:class:`~kombu.Queue` s to.
        :param exchange: The `kombu.Connection` to your message broker.
        :param queue_name: Name for your queue.
        :param routing_keys: List of routing keys that you will create
            :py:class:`~kombu.Queue` s to consume from.
        :param handler_fn: A function that will process an individual message from a queue.
        :param error_handler_fn: A function that will be called when an error is thrown
            while executing the `handler_fn`. This function will be responsible for calling
            `message.ack` or `message.requeue` as it will not be automatically called by
            `KombuMessageHandler`'s `handle` function.
        :param health_check_fn: A `baseplate.server.queue_consumer.HealthcheckCallback`
            function that can be used to customize your health check.
        :param serializer: A `baseplate.clients.kombu.KombuSerializer` that should
            be used to decode the messages you are consuming.
        :param worker_kwargs: A dictionary of keyword arguments used to configure a
            queue consumer.
        """
        queues = []
        for routing_key in routing_keys:
            queues.append(kombu.Queue(name=queue_name, exchange=exchange, routing_key=routing_key))
        return cls(
            baseplate=baseplate,
            name=queue_name,
            connection=connection,
            queues=queues,
            handler_fn=handler_fn,
            error_handler_fn=error_handler_fn,
            health_check_fn=health_check_fn,
            serializer=serializer,
            worker_kwargs=worker_kwargs,
        )

    def build_pump_worker(self, work_queue: WorkQueue) -> KombuConsumerWorker:
        kwargs = self.worker_kwargs or {}
        return KombuConsumerWorker(
            connection=self.connection,
            queues=self.queues,
            work_queue=work_queue,
            serializer=self.serializer,
            **kwargs,
        )

    def build_message_handler(self) -> KombuMessageHandler:
        return KombuMessageHandler(
            self.baseplate,
            self.name,
            self.handler_fn,
            self.error_handler_fn,
        )

    def build_health_checker(self, listener: socket.socket) -> StreamServer:
        return make_simple_healthchecker(listener, callback=self.health_check_fn)



"""Extensions to the standard library `datetime` module."""
from datetime import datetime
from datetime import timezone


def datetime_to_epoch_milliseconds(dt: datetime) -> int:
    """Convert datetime object to epoch milliseconds."""
    if not dt.tzinfo:
        dt = dt.replace(tzinfo=timezone.utc)
    return int(dt.timestamp() * 1000)


def datetime_to_epoch_seconds(dt: datetime) -> int:
    """Convert datetime object to epoch seconds."""
    return datetime_to_epoch_milliseconds(dt) // 1000


def epoch_milliseconds_to_datetime(ms: int) -> datetime:
    """Convert epoch milliseconds to UTC datetime."""
    return datetime.utcfromtimestamp(ms / 1000).replace(tzinfo=timezone.utc)


def epoch_seconds_to_datetime(sec: int) -> datetime:
    """Convert epoch seconds to UTC datetime."""
    return datetime.utcfromtimestamp(sec).replace(tzinfo=timezone.utc)


def get_utc_now() -> datetime:
    """Get current offset-aware datetime which has timezone information."""
    return datetime.now(tz=timezone.utc)
